from functools import singledispatch, update_wrapper
from types import SimpleNamespace
import numpy as np
import pprint
import json
import types
from .ops import dequantize
from .ops import np_array_from_arcs


# ----------- dummy files for geopandas and geojson ----------
class GeoDataFrame(object):
    pass


class GeoSeries(object):
    pass


geopandas = SimpleNamespace()
setattr(geopandas, "GeoDataFrame", GeoDataFrame)
setattr(geopandas, "GeoSeries", GeoSeries)


class Feature(object):
    pass


class FeatureCollection(object):
    pass


geojson = SimpleNamespace()
setattr(geojson, "Feature", Feature)
setattr(geojson, "FeatureCollection", FeatureCollection)


# ----------------- topology options object ------------------
class TopoOptions(object):
    def __init__(
        self,
        object=None,
        topology=True,
        prequantize=False,
        topoquantize=False,
        presimplify=False,
        toposimplify=False,
        simplify_with="shapely",
        simplify_algorithm="dp",
        winding_order=None,
    ):
        # get all arguments
        arguments = locals()
        if arguments["object"]:
            arguments = arguments["object"]

        if "topology" in arguments:
            self.topology = arguments["topology"]
        else:
            self.topology = True

        if "prequantize" in arguments:
            self.prequantize = arguments["prequantize"]
        else:
            self.prequantize = False

        if "topoquantize" in arguments:
            self.topoquantize = arguments["topoquantize"]
        else:
            self.topoquantize = False

        if "presimplify" in arguments:
            self.presimplify = arguments["presimplify"]
        else:
            self.presimplify = False

        if "toposimplify" in arguments:
            self.toposimplify = arguments["toposimplify"]
        else:
            self.toposimplify = False

        if "simplify_with" in arguments:
            self.simplify_with = arguments["simplify_with"]
        else:
            self.simplify_with = "shapely"

        if "simplify_algorithm" in arguments:
            self.simplify_algorithm = arguments["simplify_algorithm"]
        else:
            self.simplify_algorithm = "dp"

        if "winding_order" in arguments:
            self.winding_order = arguments["winding_order"]
        else:
            self.winding_order = None

    def __repr__(self):
        return "TopoOptions(\n  {}\n)".format(pprint.pformat(self.__dict__))


def singledispatch_class(func):
    """
    The singledispatch function only applies to functions. This function creates a
    wrapper around the singledispatch so it can be used for class instances.

    Returns
    -------
    dispatch
        dispatcher for methods
    """

    dispatcher = singledispatch(func)

    def wrapper(*args, **kw):
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)

    wrapper.register = dispatcher.register
    update_wrapper(wrapper, dispatcher)
    return wrapper


# --------- supportive functions for serialization -----------
def coordinates(arcs, tp_arcs):
    """
    Return GeoJSON coordinates for the sequence(s) of arcs.
    
    The arcs parameter may be a sequence of ints, each the index of a
    coordinate sequence within tp_arcs within the entire topology
    -- describing a line string, a sequence of such sequences
    -- describing a polygon, or a sequence of polygon arcs.
    """

    if isinstance(arcs[0], int):
        coords = np.concatenate(
            [
                tp_arcs[arc if arc >= 0 else ~arc][:: arc >= 0 or -1][i > 0 :]
                for i, arc in enumerate(arcs)
            ]
        )
        coords = coords[~np.isnan(coords).any(axis=1)].tolist()
        return coords
    elif isinstance(arcs[0], (list, tuple)):
        return list(coordinates(arc, tp_arcs) for arc in arcs)
    else:
        raise ValueError("Invalid input %s", arcs)


def geometry(obj, tp_arcs, transform=None):
    """
    Converts a topology object to a geometry object.
    
    The topology object is a dict with 'type' and 'arcs' items, such as
    -- {'type': "LineString", 'arcs': [0, 1, 2]}
    """
    if obj["type"] == "GeometryCollection":
        geometries = [geometry(feat, tp_arcs) for feat in obj["geometries"]]
        return {"type": obj["type"], "geometries": geometries}

    if obj["type"] == "MultiPoint":
        scale = transform["scale"]
        translate = transform["translate"]
        coords = obj["coordinates"]
        point_coords = dequantize(np.array(coords), scale, translate).tolist()
        return {"type": obj["type"], "coordinates": point_coords}

    if obj["type"] == "Point":
        scale = transform["scale"]
        translate = transform["translate"]
        coords = [obj["coordinates"]]

        point_coord = dequantize(np.array(coords), scale, translate).tolist()
        return {"type": obj["type"], "coordinates": point_coord[0]}

    else:
        return {"type": obj["type"], "coordinates": coordinates(obj["arcs"], tp_arcs)}


def prettyjson(obj, indent=2, maxlinelength=80):
    """Renders JSON content with indentation and line splits/concatenations to fit maxlinelength.
    Only dicts, lists and basic types are supported"""

    # https://stackoverflow.com/a/56497521/104668

    items, _ = getsubitems(
        obj, itemkey="", islast=True, maxlinelength=maxlinelength, level=0
    )
    return indentitems(items, indent, level=0)


def getsubitems(obj, itemkey, islast, maxlinelength, level):
    items = []
    is_inline = (
        True  # at first, assume we can concatenate the inner tokens into one line
    )

    isdict = isinstance(obj, dict)
    islist = isinstance(obj, list)
    istuple = isinstance(obj, tuple)
    isbasictype = not (isdict or islist or istuple)

    # build json content as a list of strings or child lists
    if isbasictype:
        # render basic type
        keyseparator = "" if itemkey == "" else ": "
        itemseparator = "" if islast else ","
        items.append(itemkey + keyseparator + basictype2str(obj) + itemseparator)

    else:
        # render lists/dicts/tuples
        if isdict:
            opening, closing, keys = ("{", "}", iter(obj.keys()))
        elif islist:
            opening, closing, keys = ("[", "]", range(0, len(obj)))
        elif istuple:
            opening, closing, keys = (
                "[",
                "]",
                range(0, len(obj)),
            )  # tuples are converted into json arrays

        if itemkey != "":
            opening = itemkey + ": " + opening
        if not islast:
            closing += ","

        count = 0
        itemkey = ""
        subitems = []

        # get the list of inner tokens
        for (i, k) in enumerate(keys):
            islast_ = i == len(obj) - 1
            itemkey_ = ""
            if isdict:
                itemkey_ = basictype2str(k)
            inner, is_inner_inline = getsubitems(
                obj[k], itemkey_, islast_, maxlinelength, level + 1
            )
            subitems.extend(inner)  # inner can be a string or a list
            is_inline = (
                is_inline and is_inner_inline
            )  # if a child couldn't be rendered inline, then we are not able either

        # fit inner tokens into one or multiple lines, each no longer than maxlinelength
        if is_inline:
            multiline = True

            # in Multi-line mode items of a list/dict/tuple can be rendered in multiple lines if they don't fit on one.
            # suitable for large lists holding data that's not manually editable.

            # in Single-line mode items are rendered inline if all fit in one line, otherwise each is rendered in a separate line.
            # suitable for smaller lists or dicts where manual editing of individual items is preferred.

            # this logic may need to be customized based on visualization requirements:
            if isdict:
                multiline = False
            if islist:
                multiline = True

            if multiline:
                lines = []
                current_line = ""
                current_index = 0

                for (i, item) in enumerate(subitems):
                    item_text = item
                    if i < len(inner) - 1:
                        item_text = item + ","

                    if len(current_line) > 0:
                        try_inline = current_line + " " + item_text
                    else:
                        try_inline = item_text

                    if len(try_inline) > maxlinelength:
                        # push the current line to the list if maxlinelength is reached
                        if len(current_line) > 0:
                            lines.append(current_line)
                        current_line = item_text
                    else:
                        # keep fitting all to one line if still below maxlinelength
                        current_line = try_inline

                    # Push the remainder of the content if end of list is reached
                    if i == len(subitems) - 1:
                        lines.append(current_line)

                subitems = lines
                if len(subitems) > 1:
                    is_inline = False
            else:  # single-line mode
                totallength = len(subitems) - 1  # spaces between items
                for item in subitems:
                    totallength += len(item)
                if totallength <= maxlinelength:
                    str = ""
                    for item in subitems:
                        str += (
                            item + " "
                        )  # insert space between items, comma is already there
                    subitems = [str.strip()]  # wrap concatenated content in a new list
                else:
                    is_inline = False

        # attempt to render the outer brackets + inner tokens in one line
        if is_inline:
            item_text = ""
            if len(subitems) > 0:
                item_text = subitems[0]
            if len(opening) + len(item_text) + len(closing) <= maxlinelength:
                items.append(opening + item_text + closing)
            else:
                is_inline = False

        # if inner tokens are rendered in multiple lines already, then the outer brackets remain in separate lines
        if not is_inline:
            items.append(opening)  # opening brackets
            items.append(subitems)  # Append children to parent list as a nested list
            items.append(closing)  # closing brackets

    return items, is_inline


def basictype2str(obj):
    if isinstance(obj, str):
        strobj = '"' + str(obj) + '"'
    elif isinstance(obj, bool):
        strobj = {True: "true", False: "false"}[obj]
    else:
        strobj = str(obj)
    return strobj


def indentitems(items, indent, level):
    """Recursively traverses the list of json lines, adds indentation based on the current depth"""
    res = ""
    indentstr = " " * (indent * level)
    for (i, item) in enumerate(items):
        if isinstance(item, list):
            res += indentitems(item, indent, level + 1)
        else:
            islast = i == len(items) - 1
            # no new line character after the last rendered line
            if level == 0 and islast:
                res += indentstr + item
            else:
                res += indentstr + item + "\n"
    return res


# ----------------- serialization functions ------------------
def serialize_as_geodataframe(topo_object, url=False):
    """
    Convert a topology dictionary or string into a GeoDataFrame.

    Parameters
    ----------
    topo_object : dict, str
        a complete object representing an topojson encoded file as
        dict, str-object or str-url

    Returns
    -------
    gdf : geopandas.GeoDataFrame
        topojson object parsed GeoDataFrame
    """
    import fiona
    import geopandas
    import json

    # parse the object as byte string
    if isinstance(topo_object, dict):
        bytes_topo = str.encode(json.dumps(topo_object))
    elif url is True:
        import requests

        request = requests.get(topo_object)
        bytes_topo = bytes(request.content)
    else:
        bytes_topo = str.encode(topo_object)
    # into an in-memory file
    vsimem = fiona.ogrext.buffer_to_virtual_file(bytes_topo)

    # read the features from a fiona collection into a GeoDataFrame
    with fiona.Collection(vsimem, driver="TopoJSON") as f:
        gdf = geopandas.GeoDataFrame.from_features(f, crs=f.crs)
    return gdf


def serialize_as_svg(topo_object, separate=False, include_junctions=False):
    from IPython.display import SVG, display
    from shapely import geometry

    keys = topo_object.keys()
    if "arcs" in keys:
        arcs = topo_object["arcs"]
        # dequantize if quantization is applied
        if "transform" in keys:

            np_arcs = np_array_from_arcs(arcs)

            transform = topo_object["transform"]
            scale = transform["scale"]
            translate = transform["translate"]

            np_arcs = dequantize(np_arcs, scale, translate)
            l_arcs = []
            for ls in np_arcs:
                l_arcs.append(ls[~np.isnan(ls)[:, 0]].tolist())
            arcs = l_arcs

        arcs = [geometry.LineString(arc) for arc in arcs]

    else:
        arcs = topo_object["linestrings"]

    if separate and not include_junctions:
        for ix, line in enumerate(arcs):
            svg = line._repr_svg_()
            print(ix, line.wkt)
            display(SVG(svg))
    elif separate and include_junctions:
        pts = topo_object["junctions"]
        for ix, line in enumerate(arcs):
            svg = geometry.GeometryCollection(
                [line, geometry.MultiPoint(pts)]
            )._repr_svg_()
            print(ix, line.wkt)
            display(SVG(svg))

    elif not separate and include_junctions:
        pts = topo_object["junctions"]
        display(
            geometry.GeometryCollection(
                [geometry.MultiLineString(arcs), geometry.MultiPoint(pts)]
            )
        )
    else:
        display(geometry.MultiLineString(arcs))


def serialize_as_json(topo_object, fp, pretty=False, indent=4, maxlinelength=88):
    if fp:
        with open(fp, "w") as f:
            if pretty:
                print(
                    prettyjson(topo_object, indent=indent, maxlinelength=maxlinelength),
                    file=f,
                )
            else:
                print(topo_object, file=f)
    else:
        if pretty:
            return prettyjson(topo_object, indent=indent, maxlinelength=maxlinelength)
        else:
            return json.dumps(topo_object)


def serialize_as_geojson(
    topo_object,
    fp=None,
    pretty=False,
    indent=4,
    maxlinelength=88,
    validate=False,
    lyr_idx=0,
):
    from shapely.geometry import asShape

    # prepare arcs from topology object
    arcs = topo_object["arcs"]
    if arcs:
        np_arcs = np_array_from_arcs(arcs)

        transform = None
        # dequantize if quantization is applied
        if "transform" in topo_object.keys():

            transform = topo_object["transform"]
            scale = transform["scale"]
            translate = transform["translate"]

            np_arcs = dequantize(np_arcs, scale, translate)

    # select object member from topology object
    lyr_name = list(topo_object["objects"].keys())[lyr_idx]
    features = topo_object["objects"][lyr_name]["geometries"]

    # prepare geojson featurecollection
    fc = {"type": "FeatureCollection", "features": []}

    # fill the featurecollection with geometry object members
    for index, feature in enumerate(features):
        f = {"id": index, "type": "Feature"}
        if "properties" in feature.keys():
            f["properties"] = feature["properties"].copy()

        # the transform is only used in cases of points or multipoints
        geommap = geometry(feature, np_arcs, transform)
        if validate:
            geom = asShape(geommap).buffer(0)
            assert geom.is_valid
            f["geometry"] = geom.__geo_interface__
        else:
            f["geometry"] = geommap

        fc["features"].append(f)
    return fc


def serialize_as_altair(
    topo_object,
    mesh=True,
    color=None,
    tooltip=True,
    projection="identity",
    objectname="data",
):
    import altair as alt

    # create a mesh visualization
    if mesh and color is None:
        data = alt.InlineData(
            values=topo_object, format=alt.DataFormat(mesh=objectname, type="topojson")
        )
        chart = (
            alt.Chart(data, width=300)
            .mark_geoshape(filled=False)
            .project(type=projection, reflectY=True)
        )

    # creating a chloropleth visualisation
    elif color is not None:
        data = alt.InlineData(
            values=topo_object,
            format=alt.DataFormat(feature=objectname, type="topojson"),
        )
        if tooltip is True:
            tooltip = [color]
        chart = (
            alt.Chart(data, width=300)
            .mark_geoshape()
            .encode(
                color=alt.Color(color, legend=alt.Legend(columns=2)), tooltip=tooltip
            )
            .project(type=projection, reflectY=True)
        )

    return chart


def serialize_as_ipywidgets(topo_object, toposimplify, topoquantize):
    from ipywidgets import interact
    from ipywidgets import fixed
    import ipywidgets as widgets

    style = {"description_width": "initial"}
    ts = toposimplify
    tq = topoquantize

    # set to simplification package for speed
    topo_object.output["options"].simplify_with = "simplification"

    alg = widgets.RadioButtons(
        options=[("Douglas-Peucker", "dp"), ("Visvalingam-Whyatt", "vw")],
        value="dp",
        description="Simplify algortihm",
        disabled=False,
        style=style,
    )
    eps = widgets.FloatSlider(
        min=ts["min"],
        max=ts["max"],
        step=ts["step"],
        value=ts["value"],
        description="Toposimplify Factor",
        style=style,
    )
    qnt = widgets.FloatLogSlider(
        min=tq["min"],
        max=tq["max"],
        step=tq["step"],
        value=tq["value"],
        base=tq["base"],
        description="Topoquantize Factor",
        style=style,
    )

    return interact(
        toposimpquant, epsilon=eps, quant=qnt, algo=alg, topo=fixed(topo_object)
    )


def toposimpquant(epsilon, quant, algo, topo):
    topo.output["options"].simplify_algorithm = algo
    return topo.toposimplify(epsilon).topoquantize(quant).to_alt()



from shapely import geometry
import topojson
import geopandas
from topojson.core.join import Join


# the returned hashmap has true for junction points
def test_join_true_for_junction_points():
    data = {
        "cba": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
        "ab": {"type": "LineString", "coordinates": [[0, 0], [1, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.MultiPoint(topo["junctions"]).equals(
        geometry.MultiPoint([(1.0, 0.0), (0.0, 0.0), (2.0, 0.0)])
    )


# the returned hashmap has undefined for non-junction points
def test_join_undefined_for_non_junction_points():
    data = {
        "cba": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
        "ab": {"type": "LineString", "coordinates": [[0, 0], [2, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.Point(1.0, 0.0) not in geometry.MultiPoint(topo["junctions"])


# forward backward lines
def test_join_forward_backward_lines():
    data = {
        "foo": {
            "type": "LineString",
            "coordinates": [(0, 0), (10, 0), (10, 5), (20, 5)],
        },
        "bar": {
            "type": "LineString",
            "coordinates": [(5, 0), (30, 0), (30, 5), (0, 5)],
        },
    }
    topo = Join(data).to_dict()

    assert len(topo["junctions"]) == 5


# more than two lines
def test_join_more_than_two_lines():
    data = {
        "foo": {"type": "LineString", "coordinates": [(0, 0), (15, 2.5), (30, 5)]},
        "bar": {"type": "LineString", "coordinates": [(0, 0), (15, 2.5), (30, 5)]},
        "baz": {
            "type": "LineString",
            "coordinates": [(0, 0), (10, 0), (10, 5), (20, 5)],
        },
        "qux": {
            "type": "LineString",
            "coordinates": [(5, 0), (30, 0), (30, 5), (0, 5)],
        },
    }
    topo = Join(data).to_dict()

    assert len(topo["junctions"]) == 5


# exact duplicate lines ABC & ABC have no junctions
def test_join_duplicate_lines_junction_endpoints():
    data = {
        "abc1": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "abc2": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# reversed duplicate lines ABC & CBA have no junctions
def test_join_reversed_duplicate_lines_junction_endpoints():
    data = {
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "cba": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# exact duplicate rings ABCA & ABCA have no junctions
def test_join_exact_duplicate_rings():
    data = {
        "abca1": {"type": "Polygon", "coordinates": [[[0, 0], [1, 1], [2, 0], [0, 0]]]},
        "abca2": {"type": "Polygon", "coordinates": [[[0, 0], [1, 1], [2, 0], [0, 0]]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# reversed duplicate rings ABCA & ACBA have no junctions
def test_join_reversed_duplicate_rings():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 1], [2, 0], [0, 0]]]},
        "acba": {"type": "Polygon", "coordinates": [[[0, 0], [2, 0], [1, 1], [0, 0]]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# rotated duplicate rings BCAB & ABCA have no junctions
def test_join_rotated_duplicate_rings():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 1], [2, 0], [0, 0]]]},
        "bcab": {"type": "Polygon", "coordinates": [[[1, 1], [2, 0], [0, 0], [1, 1]]]},
    }
    topo = Join(data).to_dict()
    assert topo["junctions"] == []


# ring ABCA & line ABCA have no junction at A
def test_join_equal_ring_and_line_no_junctions():
    data = {
        "abcaLine": {
            "type": "LineString",
            "coordinates": [[0, 0], [1, 1], [2, 0], [0, 0]],
        },
        "abcaPolygon": {
            "type": "Polygon",
            "coordinates": [[[0, 0], [1, 1], [2, 0], [0, 0]]],
        },
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# ring ABCA & line ABCA have no junctions
def test_join_rotated_ring_and_line_no_junctions():
    data = {
        "abcaLine": {
            "type": "LineString",
            "coordinates": [[0, 0], [1, 1], [2, 0], [0, 0]],
        },
        "bcabPolygon": {
            "type": "Polygon",
            "coordinates": [[[1, 1], [2, 0], [0, 0], [1, 1]]],
        },
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# when an old arc ABC extends a new arc AB, there is a junction at B
def test_join_line_ABC_extends_new_line_AB():
    data = {
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "ab": {"type": "LineString", "coordinates": [[0, 0], [1, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.MultiPoint(topo["junctions"]).equals(
        geometry.MultiPoint([(0.0, 0.0), (1.0, 0.0)])
    )


# when a reversed old arc CBA extends a new arc AB, there is a junction at B
def test_join_reversed_line_CBA_extends_new_line_AB():
    data = {
        "cba": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
        "ab": {"type": "LineString", "coordinates": [[0, 0], [1, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.Point(1.0, 0.0) in geometry.MultiPoint(topo["junctions"])


# when a new arc ADE shares its start with an old arc ABC, there is no junction at A
def test_join_line_ADE_share_starts_with_ABC():
    data = {
        "ade": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 1], [2, 1]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# ring ABA has no junctions
def test_join_single_ring_ABCA():
    data = {
        "abca": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [1, 1], [0, 0]]}
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# ring AA is not a proper polygon geometry.
def test_join_single_ring_AA():
    data = {"aa": {"type": "Polygon", "coordinates": [[0, 0], [0, 0]]}}
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# when a new line DEC shares its end with an old line ABC, there is no junction at C
def test_join_line_DEC_share_line_ABC():
    data = {
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "dec": {"type": "LineString", "coordinates": [[0, 1], [1, 1], [2, 0]]},
    }
    topo = Join(data).to_dict()
    assert topo["junctions"] == []


# when a new line ABC extends an old line AB, there is a junction at B
def test_join_line_ABC_extends_line_AB():
    data = {
        "ab": {"type": "LineString", "coordinates": [[0, 0], [1, 0]]},
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.MultiPoint(topo["junctions"]).equals(
        geometry.MultiPoint([(0.0, 0.0), (1.0, 0.0)])
    )


# when a new line ABC extends a reversed old line BA, there is a junction at B
def test_join_line_ABC_extends_line_BA():
    data = {
        "ba": {"type": "LineString", "coordinates": [[1, 0], [0, 0]]},
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.MultiPoint(topo["junctions"]).equals(
        geometry.MultiPoint([(1.0, 0.0), (0.0, 0.0)])
    )


# when a new line starts BC in the middle of an old line ABC, there is a
# junction at B
def test_join_line_ABC_extends_line_BC():
    data = {
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "bc": {"type": "LineString", "coordinates": [[1, 0], [2, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.MultiPoint(topo["junctions"]).equals(
        geometry.MultiPoint([(1.0, 0.0), (0.0, 0.0), (2.0, 0.0)])
    )


# when a new line BC starts in the middle of a reversed old line CBA, there is a
# junction at B
def test_join_line_BC_start_middle_reversed_line_CBA():
    data = {
        "cba": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
        "bc": {"type": "LineString", "coordinates": [[1, 0], [2, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.MultiPoint(topo["junctions"]).equals(
        geometry.MultiPoint([(2.0, 0.0), (1.0, 0.0)])
    )


# when a new line ABD deviates from an old line ABC, there is a junction at B
def test_join_line_ABD_deviates_line_ABC():
    data = {
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "abd": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [3, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.MultiPoint(topo["junctions"]).equals(
        geometry.MultiPoint([(0.0, 0.0), (2.0, 0.0)])
    )


# when a new line ABD deviates from a reversed old line CBA, there is a
# junction at B
def test_join_line_ABD_deviates_line_CBA():
    data = {
        "cba": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
        "abd": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [3, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.MultiPoint(topo["junctions"]).equals(
        geometry.MultiPoint([(2.0, 0.0), (0.0, 0.0)])
    )


# when a new line DBC merges into an old line ABC, there is a junction at B
def test_join_line_DBC_merge_line_ABC():
    data = {
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "dbc": {"type": "LineString", "coordinates": [[3, 0], [1, 0], [2, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.MultiPoint(topo["junctions"]).equals(
        geometry.MultiPoint([(1.0, 0.0), (2.0, 0.0), (3.0, 0.0), (0.0, 0.0)])
    )


# when a new line DBC merges into a reversed old line CBA, there is a junction at B
def test_join_line_DBC_merge_reversed_line_CBA():
    data = {
        "cba": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
        "dbc": {"type": "LineString", "coordinates": [[3, 0], [1, 0], [2, 0]]},
    }
    topo = Join(data).to_dict()

    assert geometry.MultiPoint(topo["junctions"]).equals(
        geometry.MultiPoint([(2.0, 0.0), (1.0, 0.0), (3.0, 0.0)])
    )


# when a new line DBE shares a single midpoint with an old line ABC, there is
# no junction at B
def test_join_line_DBE_share_singe_midpoint_line_ABC():
    data = {
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "dbe": {"type": "LineString", "coordinates": [[0, 1], [1, 0], [2, 1]]},
    }
    topo = Join(data).to_dict()
    assert topo["junctions"] == []


# when a new line ABDE skips a point with an old line ABCDE, there is a no junction
def test_join_line_ABDE_skips_point_line_ABCDE():
    data = {
        "abcde": {
            "type": "LineString",
            "coordinates": [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0]],
        },
        "abde": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [3, 0], [4, 0]]},
    }
    topo = Join(data).to_dict()
    assert topo["junctions"] == []


# when a new line ABDE skips a point with a reversed old line EDCBA, there is
# no junction
def test_join_line_ABDE_skips_point_reversed_line_EDCBA():
    data = {
        "edcba": {
            "type": "LineString",
            "coordinates": [[4, 0], [3, 0], [2, 0], [1, 0], [0, 0]],
        },
        "abde": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [3, 0], [4, 0]]},
    }
    topo = Join(data).to_dict()
    assert topo["junctions"] == []


# when a line ABCDBE self-intersects with its middle, there are no junctions
def test_join_line_ABCDBE_self_intersects_with_middle():
    data = {
        "abcdbe": {
            "type": "LineString",
            "coordinates": [[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]],
        }
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# when a line ABACD self-intersects with its start, there are no junctions
def test_join_line_ABACD_self_intersects_with_middle():
    data = {
        "abacd": {
            "type": "LineString",
            "coordinates": [[0, 0], [1, 0], [0, 0], [3, 0], [4, 0]],
        }
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# when a line ABCDBD self-intersects with its end, there are no junctions
def test_join_line_ABCDBD_self_intersects_with_end():
    data = {
        "abcdbd": {
            "type": "LineString",
            "coordinates": [[0, 0], [1, 0], [4, 0], [3, 0], [4, 0]],
        }
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# when an old line ABCDBE self-intersects and shares a point B, there is
# no junction at B
def test_join_line_ABCDB_self_intersects():
    data = {
        "abcdbe": {
            "type": "LineString",
            "coordinates": [[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]],
        },
        "fbg": {"type": "LineString", "coordinates": [[0, 1], [1, 0], [2, 1]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# when a line ABCA is closed, there is no junction at A
def test_join_line_ABCA_is_closed():
    data = {
        "abca": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [0, 1], [0, 0]]}
    }
    topo = Join(data).to_dict()
    assert topo["junctions"] == []


# when a ring ABCA is closed, there are no junctions
def test_join_ring_ABCA_is_closed():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]]}
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# exact duplicate rings ABCA & ABCA share the arc ABCA, but contain no junctions
def test_join_exact_duplicate_rings_ABCA_ABCA_share_ABCA():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]]},
        "abca2": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# reversed duplicate rings ABCA & ACBA share the arc ABCA, but contain no juctions
def test_join_exact_duplicate_rings_ABCA_ACBA_share_ABCA():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]]},
        "acba": {"type": "Polygon", "coordinates": [[[0, 0], [0, 1], [1, 0], [0, 0]]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# coincident rings ABCA & BCAB share the arc BCAB, but contain no junctinos
# this is a problem though as they are considered equal in a MultiPolygon geometry.
# test will pass, but coincident rings should be rotated before dedup
def test_join_coincident_rings_ABCA_BCAB_share_BCAB():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]]},
        "bcab": {"type": "Polygon", "coordinates": [[[1, 0], [0, 1], [0, 0], [1, 0]]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# coincident rings ABCA & BACB share the arc BCAB, but contain no junctions
def test_join_coincident_rings_ABCA_BACB_share_BCAB():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]]},
        "bacb": {"type": "Polygon", "coordinates": [[[1, 0], [0, 0], [0, 1], [1, 0]]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# coincident rings ABCA & DBED share the point B, but is no junction
def test_join_coincident_rings_ABCA_DBED_share_B():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]]},
        "dbed": {"type": "Polygon", "coordinates": [[[2, 1], [1, 0], [2, 2], [2, 1]]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# coincident ring ABCA & line DBE share the point B
def test_join_coincident_ring_ABCA_and_line_DBE_share_B():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]]},
        "dbe": {"type": "LineString", "coordinates": [[2, 1], [1, 0], [2, 2]]},
    }
    topo = Join(data).to_dict()

    assert topo["junctions"] == []


# should keep junctions from partly shared paths
# this test was added since the a shared path of ABC and another shared path of ABD
# only kept the junctions at A, C and D, but not at B.
def test_join_shared_junctions_in_shared_paths():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[
        (data.name == "Togo") | (data.name == "Benin") | (data.name == "Burkina Faso")
    ]
    topo = Join(data).to_dict()

    assert len(topo["junctions"]) == 6


# this test was added since a shared path can be detected of two linestrings where
# the shared path is partly from the start and partly of the end part of the
# linestring. If any shared path are detected also add the first coordinates of both
# linestrings to be considerd as junction. Not sure what happened if this introduces
# another problem since this coordinates may not be unique anymore, and this can be
# skipped as a junction..
def test_join_shared_segment_partly_start_partly_end_segment():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[
        (data.name == "Eritrea") | (data.name == "Ethiopia") | (data.name == "Sudan")
    ]
    topo = Join(data).to_dict()

    assert len(topo["junctions"]) == 6


def test_join_non_noded_interesection():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    topo = Join(data).to_dict()  # , quant_factor=1e6)

    assert len(topo["junctions"]) == 321


def test_join_super_function_extract():
    data = geometry.GeometryCollection(
        [
            geometry.Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]),
            geometry.Polygon([[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]),
        ]
    )
    topo = Join(data).to_dict()

    assert list(topo.keys()) == [
        "type",
        "linestrings",
        "coordinates",
        "bookkeeping_geoms",
        "bookkeeping_coords",
        "objects",
        "options",
        "bbox",
        "junctions",
    ]


def test_join_point():
    data = [{"type": "Point", "coordinates": [0.5, 0.5]}]
    topo = Join(data).to_dict()

    assert topo["bbox"] == (0.5, 0.5, 0.5, 0.5)


def test_join_prequantize_points():
    data = [
        {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]},
        {"type": "Point", "coordinates": [-0.5, 1.5]},
    ]
    topo = Join(data, options={"prequantize": True}).to_dict()

    assert topo["bbox"] == (-0.5, 0.0, 1.0, 1.5)


def test_join_linemerge_multilinestring():
    data = [
        {"type": "LineString", "coordinates": [(0, 0), (10, 0), (10, 5), (20, 5)]},
        {
            "type": "LineString",
            "coordinates": [
                (5, 0),
                (25, 0),
                (25, 5),
                (16, 5),
                (16, 10),
                (14, 10),
                (14, 5),
                (0, 5),
            ],
        },
    ]
    topo = Join(data).to_dict()

    assert len(topo["linestrings"]) == 2
    assert len(topo["junctions"]) == 7



import itertools
import numpy as np
from shapely import geometry
from shapely import wkt
from shapely.strtree import STRtree
import pprint
import copy
import logging


def asvoid(arr):
    """
    Utility function to create a 1-dimensional numpy void object (bytes)
    of a 2-dimensional array. This is useful for the function numpy.in1d(),
    since it only accepts 1-dimensional objects.

    Parameters
    ----------
    arr : numpy.array
        2-dimensional numpy array

    Returns
    -------
    numpy.void
        1-dimensional numpy void object
    """

    arr = np.ascontiguousarray(arr)
    if np.issubdtype(arr.dtype, np.floating):
        """ Care needs to be taken here since
        np.array([-0.]).view(np.void) != np.array([0.]).view(np.void)
        Adding 0. converts -0. to 0.
        """
        arr += 0.0
    return arr.view(np.dtype((np.void, arr.dtype.itemsize * arr.shape[-1])))


def insert_coords_in_line(line, tree_splitter):
    """
    Insert coordinates that are on the line, but where no vertices exists

    Parameters
    ----------
    line : numpy.array
        numpy array with coordinates representing a line segment
    tree_splitter: STRtree
        a STRtree splitter object

    Returns
    -------
    new_ls_xy : numpy.array
        numpy array with inserted coordinates, if any, representing a line segment
    pts_xy_on_line : numpy.array
        numpy array with with coordinates that are on the lin
    """

    # get junctions that contain within bbox line
    pts_within_bbox = tree_splitter.query(line)

    # select junctions that are within tolerance of line
    tol_dist = 1e-8
    pts_on_line = list(
        itertools.compress(
            pts_within_bbox, [line.distance(pt) < tol_dist for pt in pts_within_bbox]
        )
    )

    if len(pts_on_line) == 0:
        # no point on line, nothing to insert, nothing to split
        return None, None

    # convert shapely linestring and multipoint to np.array if there are points on line
    ls_xy = np.array(line.xy).T

    pts_xy_on_line = np.squeeze(np.array([pt.xy for pt in pts_on_line]), axis=(2,))

    # select junctions having non existing vertices in linestring
    tol_float_prc = 1e8
    pts_xy_nonexst = pts_xy_on_line[
        ~np.in1d(
            asvoid(np.around(pts_xy_on_line * tol_float_prc).astype(np.int64)),
            asvoid(np.around(ls_xy * tol_float_prc).astype(np.int64)),
        )
    ]
    if pts_xy_nonexst.size == 0:
        return ls_xy, pts_xy_on_line

    # compute the distance from the beginning of the linestring for each junction on line
    splitter_dist = np.array(
        [line.project(pt) for pt in geometry.MultiPoint(pts_xy_nonexst)]
    )
    splitter_dist = splitter_dist[splitter_dist > 0]

    # sort distance of non-existing junctions and apply sorting to the splitter distance
    # and corresponing junction coordinates to be inserted.
    sort_idx = splitter_dist.argsort()
    splitter_dist = splitter_dist[sort_idx]
    pts_xy_nonexst = pts_xy_nonexst[sort_idx]

    # get eucledian distance of all coords of line
    ls_xy_roll = np.roll(ls_xy, 1, axis=0)
    roll_min_ls = ls_xy_roll - ls_xy
    eucl_dist = np.sqrt(np.einsum("ij,ij->i", roll_min_ls, roll_min_ls))

    # the first distance is computed from the first point to the last point, set to 0
    eucl_dist[0] = 0
    ls_cumsum = eucl_dist.cumsum()

    # include junctions in linestring
    insert_idx = np.searchsorted(ls_cumsum, splitter_dist)
    new_ls_xy = np.insert(ls_xy, insert_idx, pts_xy_nonexst, axis=0)

    return new_ls_xy, pts_xy_on_line


def fast_split(line, splitter):
    """
    Split a LineString (numpy.array) with a Point or MultiPoint.
    This function is a replacement for the shapely.ops.split function, but faster.

    Parameters
    ----------
    line : numpy.array
        numpy array with coordinates that you like to be split
    splitter : numpy.array
        numpy array with coordiates on wich the line should be tried splitting

    Returns
    -------
    list of numpy.array
        If more than 1 item, the line was split. Each item in the list is a
        array of coordinates.
    """

    # previously did convert geometries of coordinates from LineString and (Multi)Point
    # to numpy arrays. This function now expect this as input to save time.
    # line = np.array(line.xy).T
    # splitter = np.squeeze(np.array([pt.xy for pt in splitter]), axis=(2,))

    # locate index of splitter coordinates in linestring
    tol = 1e8
    splitter_indices = np.flatnonzero(
        np.in1d(
            asvoid(np.around(line * tol).astype(np.int64)),
            asvoid(np.around(splitter * tol).astype(np.int64)),
        )
    )

    # compute the indices on wich to split the line
    # cannot split on first or last index of linestring
    splitter_indices = splitter_indices[
        (splitter_indices < (line.shape[0] - 1)) & (splitter_indices > 0)
    ]

    # split the linestring where each sub-array includes the split-point
    # create a new array with the index elements repeated
    tmp_indices = np.zeros(line.shape[0], dtype=np.int64)
    tmp_indices[splitter_indices] = 1
    tmp_indices += 1
    ls_xy = np.repeat(line, tmp_indices, axis=0)

    # update indices to account for the changed array
    splitter_indices = splitter_indices + np.arange(1, len(splitter_indices) + 1)

    # split using the indices as usual
    slines = np.split(ls_xy, splitter_indices, axis=0)

    return slines


def signed_area(ring):
    """
    compute the signed area of a ring (polygon)

    note: implementation is numpy variant of shapely's version:
    https://github.com/Toblerity/Shapely/blob/master/shapely/algorithms/cga.py

    Parameters
    ----------
    ring : shapely.geometry.LinearRing
        an exterior or inner ring of a shapely.geometry.Polygon

    Returns
    -------
    float
        the signed area
    """
    xs, ys = np.array(ring.coords.xy)
    signed_area = (xs * (np.roll(ys, -1) - np.roll(ys, +1))).sum() / 2
    return signed_area


def is_ccw(ring):
    """provide information if a given ring is clockwise or counterclockwise.

    Parameters
    ----------
    ring : shapely.geometry.LinearRing
        an exterior or inner ring of a shapely.geometry.Polygon

    Returns
    -------
    boolean
        True if ring is counterclockwise and False if ring is clockwise
    """
    return signed_area(ring) >= 0.0


def properties_foreign(objects):
    """
    Try to parse the object properties as foreign members. Reserved keys are:
    ["type", "bbox", "coordinates", "geometries", "geometry", "properties", "features"]

    If these keys are detected they will not be set as a foreign member and will remain
    nested within properties.

    Only if the

    Parameters
    ----------
    objects : [type]
        [description]
    """
    reserved_keys = [
        "type",
        "bbox",
        "coordinates",
        "geometries",
        "geometry",
        "properties",
        "features",
        "arcs",
    ]
    reserved_keys_used = bool(
        set(objects[0]["properties"].keys()).intersection(reserved_keys)
    )

    for obj in objects:
        reserved_keys = False
        for k, v in list(obj["properties"].items()):

            if not reserved_keys_used or k in reserved_keys:
                obj[k] = v
                obj["properties"].pop(k, None)

        if not reserved_keys_used:
            obj.pop("properties", None)

    return objects


def compare_bounds(b0, b1):
    """
    Function that compares two bounds with each other. Returns the max bound.

    Parameters
    ----------
    b0 : tuple
        tuple of xmin, ymin, xmax, ymax
    b1 : tuple
        tuple of xmin, ymin, xmax, ymax        

    Returns
    -------
    tuple
        min of mins and max of maxs
    """

    if len(b0) and len(b1):
        bounds = (
            min(b0[0], b1[0]),
            min(b0[1], b1[1]),
            max(b0[2], b1[2]),
            max(b0[3], b1[3]),
        )
    elif len(b0) and not len(b1):
        bounds = b0
    elif not len(b0) and len(b1):
        bounds = b1
    else:
        bounds = []

    return bounds


def np_array_from_lists(nested_lists):
    """
    Function to create numpy array from nested lists. The shape of the numpy array
    are the number of nested lists (rows) x the length of the longest nested list
    (columns). Rows that contain less values are filled with np.nan values.

    Parameters
    ----------
    nested_lists : list of lists
        list containing nested lists of different sizes.

    Returns
    -------
    numpy.ndarray
        array created from nested lists, np.nan is used to fill the array
    """

    np_array = np.array(list(itertools.zip_longest(*nested_lists, fillvalue=np.nan))).T
    return np_array


def lists_from_np_array(np_array):
    """
    Function to convert numpy array to list, where elements set as np.nan
    are filtered
    """

    nested_lists = [obj[~np.isnan(obj)].astype(np.int64).tolist() for obj in np_array]
    return nested_lists


def np_array_from_arcs(arcs):
    max_len_arc = len(max(arcs, key=len))
    no_arcs = len(arcs)
    np_array = np.empty((no_arcs, max_len_arc, 2))
    np_array.fill(np.nan)
    for idx in range(no_arcs):
        np_array[idx, 0 : len(arcs[idx])] = arcs[idx]
    return np_array


def dequantize(np_arcs, scale, translate):
    dequantized_arcs = np_arcs.cumsum(axis=1) * scale + translate
    return dequantized_arcs


def get_matches(geoms, tree_idx):
    """
    Function to return the indici of the rtree that intersects with the input geometries

    Parameters
    ----------
    geoms : list
        list of geometries to compare against the STRtree
    tree_idx: STRtree
        a STRtree indexing object

    Returns
    -------
    list
        list of tuples, where the key of each tuple is the linestring index and the
        value of each key is a list of junctions intersecting bounds of linestring.
    """

    # find near linestrings by querying tree
    matches = []
    for idx_ls, obj in enumerate(geoms):
        intersect_ls = tree_idx.query(obj)
        if len(intersect_ls):
            matches.extend([[[idx_ls], [ls.i for ls in intersect_ls]]])

    return matches


def select_unique(data):
    """
    Function to return unique pairs within a numpy array.
    Example: input as [[1,2], [2,1]] will return as [[1,2]]

    Parameters
    ----------
    data : numpy.array
        2 dimensional array, where each row is a couple

    Returns
    -------
    numpy.array
        2 dimensional array, where each row is unique.
    """

    sorted_data = data[np.lexsort(data.T), :]
    row_mask = np.append([True], np.any(np.diff(sorted_data, axis=0), 1))

    return sorted_data[row_mask]


def select_unique_combs(linestrings):
    """
    Given a set of inpit linestrings will create unique couple combinations.
    Each combination created contains a couple of two linestrings where the enveloppe
    overlaps each other.
    Linestrings with non-overlapping enveloppes are not returned as combination.

    Parameters
    ----------
    linestrings : list of LineString
        list where each item is a shapely LineString

    Returns
    -------
    numpy.array
        2 dimensional array, with on each row the index combination
        of a unique couple LineString with overlapping enveloppe
    """

    # create spatial index and add idx as attribute
    for i, ls in enumerate(linestrings):
        ls.i = i
    tree_idx = STRtree(linestrings)

    # get index of linestrings intersecting each linestring
    idx_match = get_matches(linestrings, tree_idx)

    # make combinations of unique possibilities
    combs = []
    for idx_comb in idx_match:
        combs.extend(list(itertools.product(*idx_comb)))

    combs = np.array(combs)
    combs.sort(axis=1)
    combs = select_unique(combs)

    uniq_line_combs = combs[(np.diff(combs, axis=1) != 0).flatten()]

    return uniq_line_combs


def quantize(linestrings, bbox, quant_factor=1e6):
    """
    Function that applies quantization. Quantization removes information by reducing
    the precision of each coordinate, effectively snapping each point to a regular grid.

    Parameters
    ----------
    linestrings: list of shapely.geometry.LineStrings
        LineStrings that will be quantized
    quant_factor : int
        Quantization factor. Normally this varies between 1e4, 1e5, 1e6. Where a
        higher number means a bigger grid where the coordinates can snap to.

    Returns
    -------
    transform : dict
        scale (kx, ky) and translation (x0, y0) values
    bbox : array
        bbox of all linestrings
    """

    x0, y0, x1, y1 = bbox
    try:
        kx = 1 / ((quant_factor - 1) / (x1 - x0))
        ky = 1 / ((quant_factor - 1) / (y1 - y0))
    except ZeroDivisionError:
        # ZeroDivisionError: float division by zero
        raise SystemExit("Cannot quantize when xmax-xmin OR ymax-ymin equals 0")

    for idx, ls in enumerate(linestrings):
        if hasattr(ls, "xy"):
            ls_xy = np.array(ls.xy)
        else:
            ls_xy = np.array(ls).T
        ls_xy = (
            np.array([(ls_xy[0] - x0) / kx, (ls_xy[1] - y0) / ky])
            .round()
            .astype(np.int64)
            .T
        )
        # get boolean slice where consecutive repeating coordinates are filtered
        bool_slice = (
            np.insert(np.absolute(np.diff(ls_xy, 1, axis=0)).sum(axis=1), 0, 1) != 0
        )
        if not bool_slice.sum() == 1 or isinstance(ls, geometry.Point):
            if isinstance(ls, geometry.Point):
                ls.coords = ls_xy[bool_slice][0]
            elif hasattr(ls, "coords"):
                ls.coords = ls_xy[bool_slice]
            else:
                linestrings[idx] = ls_xy[bool_slice].tolist()

    transform_ = {"scale": [kx, ky], "translate": [x0, y0]}

    return linestrings, transform_


def simplify(
    linestrings,
    epsilon,
    algorithm="dp",
    package="simplification",
    input_as="linestring",
):
    """
    Function that simplifies linestrings. The goal of line simplification is to reduce
    the number of points by deleting some trivial points, but without destroying the
    essential shape of the lines in the process.

    One can choose between the Douglas-Peucker ["dp"] algorithm (which simplifies
    a line based upon vertical interval) and Visvalingam–Whyatt ["vw"] (which
    progressively removes points with the least-perceptible change).

    https://pdfs.semanticscholar.org/9877/cdf50a15367bcb86649b67df8724425c5451.pdf

    Parameters
    ----------
    linestrings: list of shapely.geometry.LineStrings
        LineStrings that will be simplified
    epsilon : int
        Simplification factor. Normally this varies 1.0, 0.1 or 0.001 for "rdp" and
        30-100 for "vw".
    algorithm : str, optional
        Choose between `dp` for Douglas-Peucker and `vw` for Visvalingam–Whyatt.
        Defaults to `dp`, as its evaluation maintains to be good (Song & Miao, 2016).
    package : str, optional
        Choose between `simplification` or `shapely`. Both pachakges contains
        simplification algorithms (`shapely` only `dp`, and `simplification` both `dp`
        and `vw`).
    input_as : str, optional
        Choose between `linestring` or `array`. This function is being called from
        different locations with different input types. Choose `linestring` if the input
        type are shapely.geometry.LineString or `array` if the input are numpy.array
        coordinates

    Returns
    -------
    simp_linestrings: list of shapely.geometry.LineStrings
        LineStrings that are simplified

    Docs
    * https://observablehq.com/@lemonnish/minify-topojson-in-the-browser
    * https://github.com/topojson/topojson-simplify#planarTriangleArea
    * https://www.jasondavies.com/simplify/
    * https://bost.ocks.org/mike/simplify/
    """
    if package == "shapely":

        if input_as == "array":
            list_arcs = []
            for ls in linestrings:
                coords_to_simp = ls[~np.isnan(ls)[:, 0]]
                simple_ls = geometry.LineString(coords_to_simp)
                simple_ls = simple_ls.simplify(epsilon, preserve_topology=False)
                list_arcs.append(np.array(simple_ls).tolist())
        elif input_as == "linestring":
            for idx, ls in enumerate(linestrings):
                linestrings[idx] = ls.simplify(epsilon, preserve_topology=False)
            list_arcs = linestrings

    elif package == "simplification":
        from simplification import cutil

        if algorithm == "vw":
            alg = cutil.simplify_coords_vw

        else:
            alg = cutil.simplify_coords

        if input_as == "array":
            list_arcs = []
            for ls in linestrings:
                coords_to_simp = ls[~np.isnan(ls)[:, 0]]
                simple_ls = alg(coords_to_simp, epsilon)
                list_arcs.append(simple_ls.tolist())
        elif input_as == "linestring":
            for ls in linestrings:
                coords_to_simp = np.array(ls)
                simple_ls = alg(coords_to_simp, epsilon)
                ls.coords = simple_ls
            list_arcs = linestrings
    else:
        raise NameError(
            "Could not recognize parameter for `simplify_with`. Choose between \
                'shapely' or 'simplification'. '{}' was given".format(
                package
            )
        )
    return list_arcs


def winding_order(geom, order="CW_CCW"):
    """
    Function that force a certain winding order on the resulting output geometries. One
    can choose between "CCW_CW" and "CW_CCW".

    "CW_CCW" implies clockwise for exterior polygons and counterclockwise for interior
    polygons (aka the geographical right-hand-rule where the right hand is in the area
    of interest as you walk the line).

    "CCW_CW" implies counterclockwise for exterior polygons and clockwise for interior
    polygons (aka the mathematical right-hand-rule where the right hand curls around
    the polygon's exterior with your thumb pointing "up" (toward space), signing a
    positive area for the polygon in the signed area sense).

    TopoJSON, and so this package, defaults to "CW_CCW"*, but depending on the
    application you might decide differently.

    * https://bl.ocks.org/mbostock/a7bdfeb041e850799a8d3dce4d8c50c8

    Only applies to Polygons and MultiPolygons.

    Parameters
    ----------
    geom : geometry or shapely.geometry.GeometryCollection
        Geometry objects where the winding order will be forced upon.
    order : str, optional
        Choose "CW_CCW" for clockwise for exterior- and counterclockwise for
        interior polygons or "CCW_CW" for counterclockwise for exterior- and clockwise
        for interior polygons, by default "CW_CCW".

    Returns
    -------
    geom : geometry or shapely.geometry.GeometryCollection
        Geometry objects where the chosen winding order is forced upon.
    """

    # CW_CWW will orient the outer polygon clockwise and the inner polygon counter-
    # clockwise to conform TopoJSON standard
    if order == "CW_CCW":
        geom = geometry.polygon.orient(geom, sign=-1.0)
    elif order == "CCW_CW":
        geom = geometry.polygon.orient(geom, sign=1.0)
    else:
        raise NameError("parameter {} was not recognized".format(order))

    return geom


def round_coordinates(linestrings, rounding_precision):
    """
    Round all coordinates to a specified precision, e.g. rounding_precision=3 will round
    to 3 decimals on the resulting output geometries (after the topology is computed).

    Parameters
    ----------
    linestrings: list of shapely.geometry.LineStrings
        LineStrings of which the coordinates will be rounded
    rounding_precision : int
        Precision value. Up till how many decimales the coordinates should be rounded.

    Returns
    -------
    linestrings: list of shapely.geometry.LineStrings
        LineStrings of which the coordinates will be rounded
    """
    for idx, geom in enumerate(linestrings):
        linestrings[idx] = wkt.loads(
            wkt.dumps(geom, rounding_precision=rounding_precision)
        )
    return linestrings


def prettify(topojson_object):
    """
    prettify TopoJSON Format output for readability.

    Parameters
    ----------
    topojson_object : topojson.Topojson
        [description]

    Returns
    -------
    topojson_object : topojson.Topojson
        pretty printed JSON variant of the topoloy object
    """
    return pprint.pprint(topojson_object)


def properties_level(topojson_object, position="nested"):
    """
    Define where the attributes of the geometry object should be placed. Choose between
    "nested" or "foreign. Default is "nested" where the attribute information is placed
    within the "properties" ditctionary, part of the geometry.
    "foreign", tries to place the attributes on the same level as the geometry.

    Parameters
    ----------
    topojson_object : topojson.Topojson
        [description]
    position : str, optional
        [description], by default "nested"
    """

    import warnings

    warnings.warn(("\nNot yet implemened."), DeprecationWarning, stacklevel=2)


def delta_encoding(linestrings):
    """
    Function to apply delta-encoding to linestrings. Delta-encoding is a technique ..

    Replace in Hashmapper class.

    Parameters
    ----------
    linestrings : list of shapely.geometry.LineStrings
        LineStrings that will be delta-encoded

    Returns
    -------
    linestrings : list of shapely.geometry.LineStrings
        LineStrings that are delta-encoded
    """

    for idx, ls in enumerate(linestrings):
        ls = np.array(ls).astype(np.int64)
        ls_p1 = copy.copy(ls[0])
        ls -= np.roll(ls, 1, axis=0)
        ls[0] = ls_p1
        linestrings[idx] = ls.tolist()

    return linestrings



# pylint: disable=unsubscriptable-object
import copy
import pprint
from shapely import geometry
from shapely.wkb import loads
from shapely.ops import shared_paths
from shapely.ops import linemerge
from shapely import speedups
from ..ops import select_unique_combs
from ..ops import simplify
from ..ops import quantize
from ..ops import compare_bounds
from ..utils import serialize_as_svg
from .extract import Extract

if speedups.available:
    speedups.enable()


class Join(Extract):
    """
    This class targets the following objectives:
    1. Quantization of input linestrings if necessary
    2. Identifies junctions of shared paths

    The join function is the second step in the topology computation.
    The following sequence is adopted:
    1. extract
    2. join
    3. cut
    4. dedup
    5. hashmap

    Parameters
    ----------
    data : dict
        object created by the method topojson.extract.
    quant_factor : int, optional (default: None)
        quantization factor, used to constrain float numbers to integer values.
        - Use 1e4 for 5 valued values (00001-99999)
        - Use 1e6 for 7 valued values (0000001-9999999)

    Returns
    -------
    dict
        object expanded with
        - new key: junctions
        - new key: transform (if quant_factor is not None)
    """

    def __init__(self, data, options={}):
        # execute previous step
        super().__init__(data, options)

        # initation topology items
        self.junctions = []
        self.segments = []
        self.valerr = False

        # execute main function
        self.output = self.joiner(self.output)

    def __repr__(self):
        return "Join(\n{}\n)".format(pprint.pformat(self.output))

    def to_dict(self):
        topo_object = copy.copy(self.output)
        topo_object["options"] = vars(topo_object["options"])
        return topo_object

    def to_svg(self, separate=False, include_junctions=False):
        serialize_as_svg(self.output, separate, include_junctions)

    def joiner(self, data):
        """
        Entry point for the class Join. This function identiefs junctions
        (intersection points) of shared paths.

        The join function is the second step in the topology computation.
        The following sequence is adopted:
        1. extract
        2. join
        3. cut
        4. dedup
        5. hashmap

        Detects the junctions of shared paths from the specified hash of linestrings.

        After decomposing all geometric objects into linestrings it is necessary to
        detect the junctions or start and end-points of shared paths so these paths can
        be 'merged' in the next step. Merge is quoted as in fact only one of the
        shared path is kept and the other path is removed.

        Parameters
        ----------
        data : dict
            object created by the method topojson.extract.
        quant_factor : int, optional (default: None)
            quantization factor, used to constrain float numbers to integer values.
            - Use 1e4 for 5 valued values (00001-99999)
            - Use 1e6 for 7 valued values (0000001-9999999)

        Returns
        -------
        dict
            object expanded with
            - new key: junctions
            - new key: transform (if quant_factor is not None)
        """

        # presimplify linestrings if required
        if self.options.presimplify > 0:
            # set default if not specifically given in the options
            if type(self.options.presimplify) == bool:
                simplify_factor = 2
            else:
                simplify_factor = self.options.presimplify

            data["linestrings"] = simplify(
                data["linestrings"],
                simplify_factor,
                algorithm=self.options.simplify_algorithm,
                package=self.options.simplify_with,
                input_as="linestring",
            )

        # compute the bounding box of input geometry
        lsbs = geometry.asMultiLineString(data["linestrings"]).bounds
        ptbs = geometry.asMultiPoint(data["coordinates"]).bounds
        data["bbox"] = compare_bounds(lsbs, ptbs)

        # prequantize linestrings if required
        if self.options.prequantize > 0:
            # set default if not specifically given in the options
            if type(self.options.prequantize) == bool:
                quant_factor = 1e6
            else:
                quant_factor = self.options.prequantize

            data["linestrings"], data["transform"] = quantize(
                data["linestrings"], data["bbox"], quant_factor
            )

            data["coordinates"], data["transform"] = quantize(
                data["coordinates"], data["bbox"], quant_factor
            )

        if not self.options.topology or not data["linestrings"]:
            data["junctions"] = self.junctions
            return data

        # create list with unique combinations of lines using a rdtree
        line_combs = select_unique_combs(data["linestrings"])

        # iterate over index combinations
        for i1, i2 in line_combs:
            g1 = data["linestrings"][i1]
            g2 = data["linestrings"][i2]

            # check if geometry are equal
            # being equal meaning the geometry object coincide with each other.
            # a rotated polygon or reversed linestring are both considered equal.
            if not g1.equals(g2):
                # geoms are unique, let's find junctions
                self.shared_segs(g1, g2)

        # self.segments are nested lists of LineStrings, get coordinates of each nest
        s_coords = []
        for segment in self.segments:
            s_coords.extend(
                [
                    [
                        (x.xy[0][y], x.xy[1][y])
                        for x in segment
                        for y in range(len(x.xy[0]))
                    ]
                ]
            )
            # s_coords.extend([[y for x in segment for y in list(x.coords)]])

        # only keep junctions that appear only once in each segment (nested list)
        # coordinates that appear multiple times are not junctions
        for coords in s_coords:
            self.junctions.extend(
                [geometry.Point(i) for i in coords if coords.count(i) == 1]
            )

        # junctions can appear multiple times in multiple segments, remove duplicates
        self.junctions = [
            loads(xy) for xy in list(set([x.wkb for x in self.junctions]))
        ]

        # prepare to return object
        data["junctions"] = self.junctions

        return data

    def validate_linemerge(self, merged_line):
        """
        Return list of linestrings. If the linemerge was a MultiLineString 
        then returns a list of multiple single linestrings
        """

        if not isinstance(merged_line, geometry.LineString):
            merged_line = [ls for ls in merged_line]
        else:
            merged_line = [merged_line]
        return merged_line

    def shared_segs(self, g1, g2):
        """
        This function returns the segments that are shared with two input geometries.
        The shapely function `shapely.ops.shared_paths()` is adopted and can catch
        both the shared paths with the same direction for both inputs as well as the
        shared paths with the opposite direction for one the two inputs.

        The returned object extents the `segments` property with detected segments.
        Where each seperate segment is a linestring between two points.

        Parameters
        ----------
        g1 : shapely.geometry.LineString
            first geometry
        g2 : shapely.geometry.LineString
            second geometry
        """

        # detect potential shared paths between two linestrings
        try:
            fw_bw = shared_paths(g1, g2)
        except ValueError:
            self.valerr = True
            fw_bw = False
            # fw_bw = shared_paths(snap(g1, g2, tolerance=6), g2)

        # continue if any shared path was detected
        if fw_bw and not fw_bw.is_empty:

            forward = fw_bw[0]
            backward = fw_bw[1]

            if backward.is_empty:
                # only contains forward objects
                shared_segments = forward
            elif forward.is_empty:
                # only contains backward objects
                shared_segments = backward
            else:
                # both backward and forward contains objects, so combine
                forward = self.validate_linemerge(linemerge(forward))
                backward = self.validate_linemerge(linemerge(backward))

                shared_segments = geometry.MultiLineString(forward + backward)

            # add shared paths to segments
            self.segments.extend([list(shared_segments)])

            # also add the first coordinates of both geoms as a vertice to segments
            p1_g1 = geometry.Point([g1.xy[0][0], g1.xy[1][0]])
            p1_g2 = geometry.Point([g2.xy[0][0], g2.xy[1][0]])
            ls_p1_g1g2 = geometry.LineString([p1_g1, p1_g2])
            self.segments.extend([[ls_p1_g1g2]])



import json
from topojson.core.extract import Extract
from shapely import geometry
import geopandas
import geojson


# extract copies coordinates sequentially into a buffer
def test_extract_linestring():
    data = {
        "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "bar": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
    }
    topo = Extract(data).to_dict()

    assert len(topo["linestrings"]) == 2


# assess if a multipolygon with hole is processed into the right number of rings
def test_extract_multipolygon():
    # multipolygon with hole
    data = {
        "foo": {
            "type": "MultiPolygon",
            "coordinates": [
                [
                    [[0, 0], [20, 0], [10, 20], [0, 0]],  # CCW
                    [[3, 2], [10, 16], [17, 2], [3, 2]],  # CW
                ],
                [[[6, 4], [14, 4], [10, 12], [6, 4]]],  # CCW
                [[[25, 5], [30, 10], [35, 5], [25, 5]]],
            ],
        }
    }
    topo = Extract(data).to_dict()

    assert len(topo["bookkeeping_geoms"]) == 3
    assert len(topo["linestrings"]) == 4


# a LineString without coordinates is an empty polygon geometry
def test_extract_empty_linestring():
    data = {"empty_ls": {"type": "LineString", "coordinates": None}}
    topo = Extract(data).to_dict()

    assert topo["objects"]["empty_ls"]["arcs"] == None


# invalid polygon geometry
def test_extract_invalid_polygon():
    data = {
        "wrong": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [2, 0], [0, 0]]]},
        "valid": {"type": "Polygon", "coordinates": [[[0, 0], [2, 0], [1, 1], [0, 0]]]},
    }
    topo = Extract(data).to_dict()

    assert len(topo["linestrings"]) == 1


# test multiliinestring
def test_extract_multilinestring():
    data = {
        "foo": {
            "type": "MultiLineString",
            "coordinates": [
                [[0.0, 0.0], [1, 1], [3, 3]],
                [[1, 1], [0, 1]],
                [[3, 3], [4, 4], [0, 1]],
            ],
        }
    }
    topo = Extract(data).to_dict()

    assert len(topo["linestrings"]) == 3


# test nested geojosn geometrycollection collection
def test_extract_nested_geometrycollection():
    data = {
        "foo": {
            "type": "GeometryCollection",
            "geometries": [
                {
                    "type": "GeometryCollection",
                    "geometries": [
                        {"type": "LineString", "coordinates": [[0.1, 0.2], [0.3, 0.4]]}
                    ],
                },
                {
                    "type": "Polygon",
                    "coordinates": [[[0.5, 0.6], [0.7, 0.8], [0.9, 1.0]]],
                },
            ],
        }
    }
    topo = Extract(data).to_dict()

    assert len(topo["objects"]["foo"]["geometries"][0]["geometries"][0]["arcs"]) == 1


# test geometry collection + polygon
def test_extract_geometrycollection_polygon():
    data = {
        "bar": {"type": "Polygon", "coordinates": [[[0, 0], [1, 1], [2, 0]]]},
        "foo": {
            "type": "GeometryCollection",
            "geometries": [
                {"type": "LineString", "coordinates": [[0.1, 0.2], [0.3, 0.4]]}
            ],
        },
    }
    topo = Extract(data).to_dict()

    assert len(topo["linestrings"]) == 2


# test feature type
def test_extract_features():
    data = {
        "foo": {
            "type": "Feature",
            "geometry": {"type": "LineString", "coordinates": [[0.1, 0.2], [0.3, 0.4]]},
        },
        "bar": {
            "type": "Feature",
            "geometry": {
                "type": "Polygon",
                "coordinates": [[[0.5, 0.6], [0.7, 0.8], [0.9, 1.0]]],
            },
        },
    }
    topo = Extract(data).to_dict()

    assert len(topo["linestrings"]) == 2


# test feature collection including geometry collection
def test_extract_featurecollection():
    data = {
        "collection": {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "LineString",
                        "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                    },
                },
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "GeometryCollection",
                        "geometries": [
                            {
                                "type": "Polygon",
                                "coordinates": [[[0.5, 0.6], [0.7, 0.8], [0.9, 1.0]]],
                            }
                        ],
                    },
                },
            ],
        }
    }
    topo = Extract(data).to_dict()

    assert len(topo["objects"]) == 2
    assert len(topo["bookkeeping_geoms"]) == 2
    assert len(topo["linestrings"]) == 2
    assert topo["objects"]["feature_0"]["geometries"][0]["type"] == "LineString"
    assert (
        topo["objects"]["feature_1"]["geometries"][0]["geometries"][0]["type"]
        == "Polygon"
    )


# test to parse feature collection from a geojson file through geojson library
def test_extract_geojson_feat_col_geom_col():
    with open("tests/files_geojson/feature_collection.geojson") as f:
        data = geojson.load(f)
    topo = Extract(data).to_dict()

    assert len(topo["objects"]) == 1
    assert len(topo["bookkeeping_geoms"]) == 3
    assert len(topo["linestrings"]) == 3


# test to parse a feature from a geojson file through geojson library
def test_extract_geojson_feature_geom_col():
    with open("tests/files_geojson/feature.geojson") as f:
        data = geojson.load(f)
    topo = Extract(data).to_dict()

    assert len(topo["objects"]) == 1
    assert len(topo["bookkeeping_geoms"]) == 3
    assert len(topo["linestrings"]) == 3


# test feature collection including geometry collection
def test_extract_geopandas_geoseries():
    data = geopandas.GeoSeries(
        [
            geometry.Polygon([(0, 0), (1, 0), (1, 1)]),
            geometry.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]),
            geometry.Polygon([(2, 0), (3, 0), (3, 1), (2, 1)]),
        ]
    )
    topo = Extract(data).to_dict()

    assert len(topo["objects"]) == 3
    assert len(topo["bookkeeping_geoms"]) == 3
    assert len(topo["linestrings"]) == 3
    # TEST FAILS because of https://github.com/geopandas/geopandas/issues/1070


# test shapely geometry collection.
def test_extract_shapely_geometrycollection():
    data = geometry.GeometryCollection(
        [
            geometry.Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]),
            geometry.Polygon([[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]),
        ]
    )
    topo = Extract(data).to_dict()

    assert len(topo["objects"]) == 1
    assert len(topo["bookkeeping_geoms"]) == 2
    assert len(topo["linestrings"]) == 2


def test_extract_geo_interface_from_list():
    data = [
        {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
    ]
    topo = Extract(data).to_dict()

    assert len(topo["linestrings"]) == 2


def test_extract_shapely_geo_interface_from_list():
    data = [
        geometry.Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]),
        geometry.Polygon([[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]),
    ]
    topo = Extract(data).to_dict()

    assert len(topo["linestrings"]) == 2
    assert isinstance(topo["objects"][0], dict)


# duplicate rotated geometry bar with hole interior in geometry foo
def test_extract_geomcol_multipolygon_polygon():
    data = {
        "foo": {
            "type": "GeometryCollection",
            "geometries": [
                {
                    "type": "MultiPolygon",
                    "coordinates": [
                        [
                            [[10, 20], [20, 0], [0, 0], [3, 13], [10, 20]],
                            [[3, 2], [10, 16], [17, 2], [3, 2]],
                        ],
                        [[[10, 4], [14, 4], [10, 12], [10, 4]]],
                    ],
                },
                {
                    "type": "Polygon",
                    "coordinates": [[[20, 0], [35, 5], [10, 20], [20, 0]]],
                },
            ],
        }
    }
    topo = Extract(data).to_dict()

    assert len(topo["linestrings"]) == 4


def test_extract_geo_interface_shapefile():
    import shapefile

    data = shapefile.Reader("tests/files_shapefile/southamerica.shp")
    topo = Extract(data).to_dict()

    assert len(topo["linestrings"]) == 15


def test_extract_points():
    data = [
        {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]},
        {"type": "Point", "coordinates": [0.5, 0.5]},
    ]
    topo = Extract(data).to_dict()

    assert len(topo["bookkeeping_coords"]) == 1
    assert len(topo["bookkeeping_geoms"]) == 1
    assert topo["coordinates"][0].wkt == "POINT (0.5 0.5)"
    assert "coordinates" in topo["objects"][1].keys()


def test_extract_single_polygon():
    data = geometry.Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]])
    topo = Extract(data).to_dict()

    assert len(topo["bookkeeping_geoms"]) == 1


def test_extract_single_linestring():
    data = geometry.LineString([[0, 0], [1, 0], [1, 1], [0, 1]])
    topo = Extract(data).to_dict()

    assert len(topo["bookkeeping_geoms"]) == 1


def test_extract_single_multilinestring():
    data = geometry.MultiLineString([[[0, 0], [1, 1]], [[-1, 0], [1, 0]]])
    topo = Extract(data).to_dict()

    assert len(topo["bookkeeping_geoms"]) == 2


def test_extract_single_multilinestring_list():
    data = [geometry.MultiLineString([[[0, 0], [1, 1]], [[-1, 0], [1, 0]]])]
    topo = Extract(data).to_dict()

    assert len(topo["bookkeeping_geoms"]) == 2



import topojson
import geopandas
from shapely import geometry
from topojson.core.hashmap import Hashmap

# duplicate rotated geometry bar with hole interior in geometry foo
def test_hashmap_geomcol_multipolygon_polygon():
    data = {
        "foo": {
            "type": "GeometryCollection",
            "geometries": [
                {
                    "type": "MultiPolygon",
                    "coordinates": [
                        [
                            [[10, 20], [20, 0], [0, 0], [10, 20]],
                            [[3, 2], [10, 16], [17, 2], [3, 2]],
                        ],
                        [[[6, 4], [14, 4], [10, 12], [6, 4]]],
                    ],
                },
                {
                    "type": "Polygon",
                    "coordinates": [[[20, 0], [35, 5], [10, 20], [20, 0]]],
                },
            ],
        }
    }
    topo = Hashmap(data).to_dict()

    assert topo["objects"]["data"]["geometries"][0]["geometries"][0]["arcs"] == [
        [[4, 0]],
        [[1]],
        [[2]],
    ]


def test_hashmap_backward_polygon():
    data = {
        "abc": {
            "type": "Polygon",
            "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]],
        },
        "def": {
            "type": "Polygon",
            "coordinates": [[[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]],
        },
    }
    topo = Hashmap(data).to_dict()

    assert topo["objects"]["data"]["geometries"][0]["arcs"] == [[-3, 0]]
    assert topo["objects"]["data"]["geometries"][1]["arcs"] == [[1, 2]]


# this test should catch a shared boundary and a hashed multipolgon
# related to https://github.com/Toblerity/Shapely/issues/535
def test_hashmap_albania_greece():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.name == "Albania") | (data.name == "Greece")]
    topo = Hashmap(data).to_dict()

    assert len(topo["linestrings"]) == 4


# something is wrong with hashmapping in the example of benin
def test_hashmap_benin_surrounding_countries():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[
        (data.name == "Togo") | (data.name == "Benin") | (data.name == "Burkina Faso")
    ]
    topo = Hashmap(data).to_dict()

    assert len(topo["linestrings"]) == 6


# something is wrong with hashmapping once a geometry has only shared arcs
def test_hashmap_geom_surrounding_many_geometries():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[
        (data.name == "Botswana")
        | (data.name == "South Africa")
        | (data.name == "Zimbabwe")
        | (data.name == "Namibia")
        | (data.name == "Zambia")
    ]
    topo = Hashmap(data).to_dict()

    assert len(topo["linestrings"]) == 13


# this test was added since the shared_arcs bookkeeping is doing well, but the
# wrong arc gots deleted. How come?
def test_hashmap_shared_arcs_ordering_issues():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[
        (data.name == "Botswana")
        | (data.name == "South Africa")
        | (data.name == "Zimbabwe")
        | (data.name == "Mozambique")
        | (data.name == "Zambia")
    ]
    topo = Hashmap(data).to_dict()
    assert len(topo["linestrings"]) == 17


def test_hashmap_super_function():
    data = geometry.GeometryCollection(
        [
            geometry.Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]),
            geometry.Polygon([[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]),
        ]
    )
    topo = Hashmap(data).to_dict()
    geoms = topo["objects"]["data"]["geometries"][0]["geometries"]

    assert list(topo.keys()) == [
        "type",
        "linestrings",
        "coordinates",
        "objects",
        "options",
        "bbox",
    ]
    assert geoms[0]["arcs"] == [[-3, 0]]
    assert geoms[1]["arcs"] == [[1, 2]]


# this test was added since objects with nested geometreycollections seems not
# being parsed in the topojson format.
# Pass for now:
def test_hashmap_of_nested_geometrycollection():
    data = {
        "foo": {
            "type": "GeometryCollection",
            "geometries": [
                {
                    "type": "GeometryCollection",
                    "geometries": [
                        {"type": "LineString", "coordinates": [[0.1, 0.2], [0.3, 0.4]]}
                    ],
                },
                {
                    "type": "Polygon",
                    "coordinates": [[[0.5, 0.6], [0.7, 0.8], [0.9, 1.0]]],
                },
            ],
        }
    }
    topo = Hashmap(data).to_dict()

    assert topo["objects"]["data"]["type"] == "GeometryCollection"


# this test was added because the winding order is still giving issues.
# see related issue: https://github.com/mattijn/topojson/issues/30
def test_hashmap_winding_order_geom_solely_shared_arcs():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[
        (data.name == "Jordan") | (data.name == "Palestine") | (data.name == "Israel")
    ]
    topo = Hashmap(data).to_dict()

    assert topo["objects"]["data"]["geometries"][1]["arcs"] == [[1, -6]]


def test_hashmap_point():
    data = [{"type": "Point", "coordinates": [0.5, 0.5]}]
    topo = Hashmap(data).to_dict()

    assert topo["bbox"] == (0.5, 0.5, 0.5, 0.5)
    assert len(topo["coordinates"]) == 1


def test_hashmap_polygon_point():
    data = [
        {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]},
        {"type": "Point", "coordinates": [0.5, 0.5]},
    ]
    topo = Hashmap(data).to_dict()

    assert len(topo["coordinates"]) == 1
    assert len(topo["linestrings"]) == 1


def test_hashmap_polygon_point():
    data = [
        {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]},
        {"type": "Point", "coordinates": [0.5, 0.5]},
    ]
    topo = Hashmap(data).to_dict()

    assert len(topo["coordinates"]) == 1
    assert len(topo["linestrings"]) == 1


def test_hashmap_multipoint():
    data = [{"type": "MultiPoint", "coordinates": [[0.5, 0.5], [1.0, 1.0]]}]
    topo = Hashmap(data).to_dict()

    assert len(topo["coordinates"]) == 2


def test_hashmap_polygon():
    data = [
        {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]}
    ]
    topo = Hashmap(data).to_dict()

    assert len(topo["linestrings"]) == 1


def test_hashmap_point_multipoint():
    data = [
        {"type": "Point", "coordinates": [0.5, 0.5]},
        {"type": "MultiPoint", "coordinates": [[0.5, 0.5], [1.0, 1.0]]},
        {"type": "Point", "coordinates": [2.5, 3.5]},
    ]
    topo = Hashmap(data).to_dict()

    assert len(topo["coordinates"]) == 4


def test_hashmap_nested_geometrycollection():
    data = {
        "collection": {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "LineString",
                        "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                    },
                },
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "GeometryCollection",
                        "geometries": [
                            {
                                "type": "Polygon",
                                "coordinates": [[[0.5, 0.6], [0.7, 0.8], [0.9, 1.0]]],
                            }
                        ],
                    },
                },
            ],
        }
    }
    topo = Hashmap(data).to_dict()

    assert len(topo["linestrings"]) == 2


def test_hashmap_polygon_geometrycollection():
    data = {
        "bar": {"type": "Polygon", "coordinates": [[[0, 0], [1, 1], [2, 0]]]},
        "foo": {
            "type": "GeometryCollection",
            "geometries": [
                {"type": "LineString", "coordinates": [[0.1, 0.2], [0.3, 0.4]]}
            ],
        },
    }
    topo = Hashmap(data).to_dict()

    assert len(topo["linestrings"]) == 2


def test_hashmap_linestring_polygon():
    data = {
        "foo": {
            "type": "Feature",
            "geometry": {"type": "LineString", "coordinates": [[0.1, 0.2], [0.3, 0.4]]},
        },
        "bar": {
            "type": "Feature",
            "geometry": {
                "type": "Polygon",
                "coordinates": [[[0.5, 0.6], [0.7, 0.8], [0.9, 1.0]]],
            },
        },
    }
    topo = Hashmap(data).to_dict()

    assert len(topo["linestrings"]) == 2


def test_hashmap_polygon_point():
    data = [
        {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]},
        {"type": "Point", "coordinates": [0.5, 0.5]},
    ]
    topo = Hashmap(data).to_dict()

    assert len(topo["linestrings"]) == 1
    assert len(topo["coordinates"]) == 1



import copy
import pprint
import numpy as np
from shapely import geometry
from shapely.ops import linemerge
from .cut import Cut
from ..ops import np_array_from_lists
from ..ops import lists_from_np_array
from ..utils import serialize_as_svg


class Dedup(Cut):
    """
    Dedup duplicates and merge contiguous arcs
    """

    def __init__(self, data, options={}):
        # execute previous step
        super().__init__(data, options)

        # initation topology items
        self.shared_arcs_idx = []
        self.merged_arcs_idx = []

        # execute main function of Dedup
        self.output = self.deduper(self.output)

    def __repr__(self):
        return "Dedup(\n{}\n)".format(pprint.pformat(self.output))

    def to_dict(self):
        topo_object = copy.copy(self.output)
        topo_object["options"] = vars(topo_object["options"])
        return topo_object

    def to_svg(self, separate=False, include_junctions=False):
        serialize_as_svg(self.output, separate, include_junctions)

    def deduper(self, data):
        """
        Deduplication of linestrings that contain duplicates

        The dedup function is the fourth step in the topology computation.
        The following sequence is adopted:
        1. extract
        2. join
        3. cut
        4. dedup
        5. hashmap
        """

        # deduplicate equal geometries
        # create numpy array from bookkeeping_geoms variable for numerical computation
        array_bk = np_array_from_lists(data["bookkeeping_linestrings"])
        array_bk_sarcs = None
        if data["bookkeeping_duplicates"].size != 0:
            array_bk_sarcs, dup_pair_list = self.deduplicate(
                data["bookkeeping_duplicates"], data["linestrings"], array_bk
            )

        # apply a shapely linemerge to merge all contiguous line-elements
        # first create a mask for shared arcs to select only non-duplicates
        mask = np.isin(array_bk, array_bk_sarcs)
        array_bk_ndp = copy.deepcopy(array_bk.astype(float))

        # only do merging of arcs if there are contigous arcs in geoms
        if array_bk_ndp[mask].size != 0:
            # make sure the idx of shared arcs are set to np.nan
            array_bk_ndp[mask] = np.nan

            # slice array_bk_ndp for geoms (rows) containing np.nan values
            slice_idx = np.all(~np.isnan(array_bk_ndp)[:, [0, -1]], axis=1)
            sliced_array_bk_ndp = array_bk_ndp[slice_idx]

            # apply linemerge on geoms containing contigious arcs and maintain
            # bookkeeping
            self.merge_contigious_arcs(data, sliced_array_bk_ndp)

            # pop the merged contigious arcs and maintain bookkeeping.
            self.pop_merged_arcs(data, array_bk, array_bk_sarcs)

        # prepare to return object
        del data["bookkeeping_linestrings"]
        data["bookkeeping_arcs"] = lists_from_np_array(array_bk)
        if data["bookkeeping_duplicates"].size != 0:
            data["bookkeeping_shared_arcs"] = array_bk_sarcs.astype(np.int64).tolist()
            data["bookkeeping_duplicates"] = lists_from_np_array(
                data["bookkeeping_duplicates"][dup_pair_list != -99]
            )
        else:
            data["bookkeeping_shared_arcs"] = []

        return data

    def find_merged_linestring(self, data, no_ndp_arcs, ndp_arcs, ndp_arcs_bk):
        """
        Function to find the index of LineString in a MultiLineString object which
        contains merged LineStrings.

        Parameters
        ----------
        data : dict
            object that contains the 'linestrings'
        no_ndp_arcs : int
            number of non-duplicate arcs
        ndp_arcs : array
            array containing index values of the related arcs

        Returns
        -------
        int
            index of LineString that contains merged LineStrings
        """

        for segment_idx in range(no_ndp_arcs):
            merged_arcs_bool = [
                ndp_arcs[segment_idx].contains(data["linestrings"][i])
                for i in ndp_arcs_bk
            ].count(True)
            if merged_arcs_bool == 2:
                return segment_idx, "first_last"
            elif merged_arcs_bool == len(ndp_arcs_bk):
                return segment_idx, "all"

    def deduplicate(self, dup_pair_list, linestring_list, array_bk):
        """
        Function to deduplicate items

        Parameters
        ----------
        dup_pair_list : numpy.ndarray
            array containing pair of indexes that refer to duplicate linestrings.
        linestring_list : list of shapely.geometry.LineStrings
            list of linestrings from which items will be removed.
        array_bk : numpy.ndarray
            array used for bookkeeping of linestrings.

        Returns
        -------
        numpy.ndarray
            bookkeeping array of shared arcs
        numpy.ndarray
            array where each processed pair is set to -99
        """

        # sort the dup_pair_list by the 1st column (idx_keep) in descending order
        dup_pair_list = dup_pair_list[dup_pair_list[:, 0].argsort()[::-1]]
        # initiate an array with np.nans, where sizes equals number of duplicates
        array_bk_sarcs = np.empty(len(dup_pair_list))
        array_bk_sarcs.fill(np.nan)

        # start deduping
        for idx, dup_pair in enumerate(dup_pair_list):
            idx_keep = dup_pair[0]
            idx_pop = dup_pair[1]

            # remove duplicate linestring
            del linestring_list[idx_pop]

            # change reference duplicate to idx_keep
            no_dups = array_bk[array_bk == idx_pop].size
            array_bk[array_bk == idx_pop] = idx_keep

            # next run may affect previous runs, maintain bookkeeping and change
            # all elements greater than the index which was removed.
            with np.errstate(invalid="ignore"):
                array_bk[array_bk > idx_pop] -= no_dups
                dup_pair_list[dup_pair_list > idx_pop] -= no_dups
                array_bk_sarcs[array_bk_sarcs > idx_pop] -= no_dups

            # store shared arc index
            idx2keep = idx_keep if idx_pop > idx_keep else idx_keep - 1
            array_bk_sarcs[idx] = idx2keep

            # set duplicate entry to -99
            dup_pair_list[idx, :] = -99

        return array_bk_sarcs, dup_pair_list

    def merge_contigious_arcs(self, data, sliced_array_bk_ndp):
        """
        Function that iterate over geoms that contain shared arcs and try linemerge
        on remaining arcs. The merged contigious arc is placed back in the 'linestrings'
        object.
        The arcs that can be popped are placed within the merged_arcs_idx list

        Parameters
        ----------
        data : dict
            object that contains the 'linestrings'.
        sliced_array_bk_ndp : numpy.ndarray
            bookkeeping array where shared linestrings are set to np.nan.
        """

        for arcs_geom_bk in sliced_array_bk_ndp:
            # set number of arcs before trying linemerge
            ndp_arcs_bk = arcs_geom_bk[~np.isnan(arcs_geom_bk)].astype(np.int64)
            no_ndp_arcs_bk = len(ndp_arcs_bk)

            # apply linemerge
            ndp_arcs = linemerge([data["linestrings"][i] for i in ndp_arcs_bk])
            if isinstance(ndp_arcs, geometry.LineString):
                ndp_arcs = [ndp_arcs]
            no_ndp_arcs = len(ndp_arcs)

            # if no_ndp_arcs is different than no_ndp_arcs_bk, than a merge took place
            # if lengths are equal, than no merge did occur and no need to solve the
            # bookkeeping
            if no_ndp_arcs != no_ndp_arcs_bk:
                # get the idx of the linestring which was merged
                idx_merg_arc, consec_behavior = self.find_merged_linestring(
                    data, no_ndp_arcs, ndp_arcs, ndp_arcs_bk
                )

                # keep last arc of non-duplicate arcs and pop the remaining arcs
                idx_keep = ndp_arcs_bk[-1]
                # replace linestring of idx_keep with merged linestring
                data["linestrings"][idx_keep] = ndp_arcs[idx_merg_arc]
                if consec_behavior == "first_last":
                    idx_pop = ndp_arcs_bk[0]
                    self.merged_arcs_idx.append(idx_pop)
                elif consec_behavior == "all":
                    idx_pop = np.delete(ndp_arcs_bk, -1)
                    self.merged_arcs_idx.extend(idx_pop.tolist())

    def pop_merged_arcs(self, data, array_bk, array_bk_sarcs):
        """
        The collected indici that can be popped, since they have been merged
        """

        self.merged_arcs_idx.sort(reverse=True)
        for idx_pop in self.merged_arcs_idx:

            # remove merged linestring
            del data["linestrings"][idx_pop]

            # change reference of merged linestring and all index elements greater
            # then idx_pop
            array_bk[array_bk == idx_pop] = np.nan
            with np.errstate(invalid="ignore"):
                array_bk[array_bk > idx_pop] -= 1
                array_bk_sarcs[array_bk_sarcs > idx_pop] -= 1



"""
topojson - a powerful library to encode geographic data as topology in Python!🌍
===============================================================================
**topojson** is a Python package with the aim of creating TopoJSON topology.

Main Features
-------------
  - Aims to create TopoJSON for _any_ geographic vector data parsed into Python
  - Ability to select the winding order of the geometric input.
  - Options to prequantize and presimplify the geometric features preparatory
    computing the topology.
  - Options to topoquantize and toposimplify after the topology is computed
  - Choose between the package `shapely` or `simplification` to simplify the
    linestrings or arcs.
  - Direct support to analyze the arcs as svg
  - Optional support to parse the TopoJSON into a GeoDataFrame if geopandas is
    installed.
  - Optional support to parse the TopoJSON directly into a mesh or geoshape in
    Altair if the package is installed.
  - Optional support for UI controls to exploring the results of topoquantize
    and toposimplify interactively if ipywidgets is installed.
"""
__version__ = "1.0rc9"
__all__ = ["Topology "]

from .core.topology import Topology  # noqa



import json
from shapely import geometry
import geopandas
import geojson
import topojson
import pytest


# this test was added since geometries of only linestrings resulted in a topojson
# file that returns an empty geodataframe when reading
def test_topology_linestrings_parsed_to_gdf():
    data = [
        {"type": "LineString", "coordinates": [[4, 0], [2, 2], [0, 0]]},
        {"type": "LineString", "coordinates": [[0, 2], [1, 1], [2, 2], [3, 1], [4, 2]]},
    ]
    topo = topojson.Topology(data).to_gdf()

    assert topo["geometry"][0].wkt != "GEOMETRYCOLLECTION EMPTY"
    assert topo["geometry"][0].type == "LineString"


# test winding order using TopoOptions object
def test_topology_winding_order_TopoOptions():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.name == "South Africa")]
    topo = topojson.Topology(data, winding_order="CW_CCW").to_dict(options=True)

    assert len(topo["objects"]) == 1
    assert len(topo["options"]) == 8


# test winding order using kwarg variables
def test_topology_winding_order_kwarg_vars():

    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.name == "South Africa")]
    topo = topojson.Topology(data, winding_order="CW_CCW").to_dict(options=True)

    assert len(topo["objects"]) == 1
    assert len(topo["options"]) == 8


def test_topology_computing_topology():
    data = [
        {"type": "LineString", "coordinates": [[4, 0], [2, 2], [0, 0]]},
        {"type": "LineString", "coordinates": [[0, 2], [1, 1], [2, 2], [3, 1], [4, 2]]},
    ]
    no_topo = topojson.Topology(data, topology=False, prequantize=False).to_dict()
    topo = topojson.Topology(data, topology=True, prequantize=False).to_dict()

    assert len(topo["arcs"]) == 5
    assert len(no_topo["arcs"]) == 2


# test prequantization without computing topology
def test_topology_prequantization():

    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[
        (data.name == "Botswana")
        | (data.name == "South Africa")
        | (data.name == "Zimbabwe")
        | (data.name == "Mozambique")
        | (data.name == "Zambia")
    ]
    topo = topojson.Topology(data, topology=False, prequantize=1e4).to_dict()

    assert "transform" in topo.keys()


# test prequantization without computing topology
def test_topology_prequantization_including_delta_encoding():

    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[
        (data.name == "Botswana")
        | (data.name == "South Africa")
        | (data.name == "Zimbabwe")
        | (data.name == "Mozambique")
        | (data.name == "Zambia")
    ]
    topo = topojson.Topology(data, topology=False, prequantize=1e4).to_dict()

    assert "transform" in topo.keys()


def test_topology_toposimplify_set_in_options():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.name == "Antarctica")]
    topo = topojson.Topology(
        data, prequantize=True, simplify_with="shapely", toposimplify=4
    ).to_dict()

    assert "transform" in topo.keys()


def test_topology_toposimplify_as_chaining():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.name == "Antarctica")]
    topo = topojson.Topology(data, prequantize=True, simplify_with="shapely")
    topos = topo.toposimplify(2).to_dict()

    assert "transform" in topos.keys()


def test_topology_topoquantize_as_chaining():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.name == "Antarctica")]
    topo = topojson.Topology(data, prequantize=False, simplify_with="shapely")
    topos = topo.topoquantize(1e2).to_dict()

    assert "transform" in topos.keys()


def test_topology_prequantize_topoquantize_as_chaining():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.name == "Antarctica")]
    topo = topojson.Topology(data, prequantize=1e6, topology=True)
    topos = topo.topoquantize(1e5).to_dict()

    assert "transform" in topos.keys()


def test_topology_to_svg():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.name == "Antarctica")]
    topo = topojson.Topology(data, prequantize=1e6, presimplify=50, topology=True)

    assert topo.to_svg() == None


def test_topology_with_arcs_without_linestrings():
    data = [
        {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]},
        {"type": "Polygon", "coordinates": [[[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]]},
    ]
    topo = topojson.Topology(data, prequantize=False, topology=True).to_dict()

    assert "linestrings" not in topo.keys()


def test_topology_widget():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.continent == "Africa")]
    topo = topojson.Topology(data, prequantize=1e6, topology=True)
    widget = topo.to_widget()

    assert len(widget.widget.children) == 4  # pylint: disable=no-member


def test_topology_simplification_vw():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.continent == "South America")]
    topo = topojson.Topology(
        data,
        prequantize=False,
        topology=True,
        toposimplify=1,
        simplify_with="simplification",
        simplify_algorithm="vw",
    ).to_dict()

    assert len(topo["arcs"][0]) == 4


def test_topology_simplification_dp():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.continent == "South America")]
    topo = topojson.Topology(
        data,
        prequantize=False,
        topology=True,
        toposimplify=1,
        simplify_with="simplification",
        simplify_algorithm="dp",
    ).to_dict()

    assert len(topo["arcs"][0]) == 3


def test_topology_polygon_point():
    data = [
        {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]},
        {"type": "Point", "coordinates": [-0.5, 1.5]},
    ]
    topo = topojson.Topology(data, topoquantize=True).to_dict()

    assert len(topo["arcs"]) == 1
    assert topo["objects"]["data"]["geometries"][1]["coordinates"] == [0, 999999]


def test_topology_point():
    data = [{"type": "Point", "coordinates": [0.5, 0.5]}]
    with pytest.raises(SystemExit) as topo:
        topojson.Topology(data, topoquantize=True).to_dict()

    assert topo.type == SystemExit
    assert topo.value.code == "Cannot quantize when xmax-xmin OR ymax-ymin equals 0"


def test_topology_multipoint():
    data = [{"type": "MultiPoint", "coordinates": [[0.5, 0.5], [1.0, 1.0]]}]
    topo = topojson.Topology(data, topoquantize=True).to_dict()

    assert len(topo["arcs"]) == 0
    assert topo["objects"]["data"]["geometries"][0]["coordinates"] == [
        [0, 0],
        [999999, 999999],
    ]
    assert topo["transform"]["translate"] == [0.5, 0.5]


def test_topology_polygon():
    data = [
        {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]}
    ]
    topo = topojson.Topology(data, topoquantize=True).to_dict()

    assert topo["transform"]["translate"] == [0.0, 0.0]


def test_topology_point_multipoint():
    data = [
        {"type": "Point", "coordinates": [0.5, 0.5]},
        {"type": "MultiPoint", "coordinates": [[0.5, 0.5], [1.0, 1.0]]},
        {"type": "Point", "coordinates": [2.5, 3.5]},
    ]
    topo = topojson.Topology(data, topoquantize=True).to_dict()

    assert topo["objects"]["data"]["geometries"][0]["coordinates"] == [0, 0]
    assert topo["objects"]["data"]["geometries"][2]["coordinates"] == [999999, 999999]


def test_topology_to_geojson_nested_geometrycollection():
    data = {
        "collection": {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "LineString",
                        "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                    },
                },
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "GeometryCollection",
                        "geometries": [
                            {
                                "type": "Polygon",
                                "coordinates": [[[0.5, 0.6], [0.7, 0.8], [0.9, 1.0]]],
                            }
                        ],
                    },
                },
            ],
        }
    }
    topo = topojson.Topology(data).to_geojson(pretty=False)

    assert "]]]}]}]}}]}" in topo


def test_topology_to_geojson_polygon_geometrycollection():
    data = {
        "bar": {"type": "Polygon", "coordinates": [[[0, 0], [1, 1], [2, 0]]]},
        "foo": {
            "type": "GeometryCollection",
            "geometries": [
                {"type": "LineString", "coordinates": [[0.1, 0.2], [0.3, 0.4]]}
            ],
        },
    }
    topo = topojson.Topology(data).to_geojson()

    assert "]]}}]}" in topo


def test_topology_to_geojson_linestring_polygon():
    data = {
        "foo": {
            "type": "Feature",
            "geometry": {"type": "LineString", "coordinates": [[0.1, 0.2], [0.3, 0.4]]},
        },
        "bar": {
            "type": "Feature",
            "geometry": {
                "type": "Polygon",
                "coordinates": [[[0.5, 0.6], [0.7, 0.8], [0.9, 1.0]]],
            },
        },
    }
    topo = topojson.Topology(data).to_geojson()

    assert "]]}}]}" in topo


def test_topology_to_geojson_polygon_point():
    data = [
        {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]},
        {"type": "Point", "coordinates": [0.5, 0.5]},
    ]
    topo = topojson.Topology(data).to_geojson()

    assert "]]]}}" in topo  # feat 1
    assert "]}}]}" in topo  # feat 2




import pprint
import copy
import numpy as np
import itertools
from .hashmap import Hashmap
from ..ops import properties_foreign
from ..ops import np_array_from_arcs
from ..ops import dequantize
from ..ops import quantize
from ..ops import simplify
from ..ops import delta_encoding
from ..utils import TopoOptions
from ..utils import serialize_as_svg
from ..utils import serialize_as_json
from ..utils import serialize_as_geojson


class Topology(Hashmap):
    """
    Returns a TopoJSON topology for the specified geometric object. TopoJSON is an 
    extension of GeoJSON providing multiple approaches to compress the geographical 
    input data. These options include simplifying the linestrings or quantizing the 
    coordinates but foremost the computation of a topology.

    Parameters
    ----------
    data : _any_ geometric type
        Geometric data that should be converted into TopoJSON 
    topology : boolean
        Specifiy if the topology should be computed for deriving the TopoJSON. 
        Default is True.
    prequantize : boolean, int
        If the prequantization parameter is specified, the input geometry is quantized
        prior to computing the topology, the returned topology is quantized, and its 
        arcs are delta-encoded. Quantization is recommended to improve the quality of 
        the topology if the input geometry is messy (i.e., small floating point error 
        means that adjacent boundaries do not have identical values); typical values 
        are powers of ten, such as 1e4, 1e5 or 1e6. Default is True (which correspond 
        to a quantize factor of 1e6).
    topoquantize : boolean or int
        If the topoquantization parameter is specified, the input geometry is quantized 
        after the topology is constructed. If the topology is already quantized this 
        will be resolved first before the topoquantization is applied. 
        Default is False.
    presimplify : boolean, float
        Apply presimplify to remove unnecessary points from linestrings before the 
        topology is constructed. This will simplify the input geometries. 
        Default is False.
    toposimplify : boolean,float 
        Apply toposimplify to remove unnecessary points from arcs after the topology is 
        constructed. This will simplify the constructed arcs without altering the 
        topological relations. Sensible values are in the range of 0.0001 to 10. 
        Defaults to 0.0001.
    simplify_with : str
        Sets the package to use for simplifying (both pre- and toposimplify). Choose 
        between `shapely` or `simplification`. Shapely adopts solely Douglas-Peucker 
        and simplification both Douglas-Peucker and Visvalingam-Whyatt. The pacakge 
        simplification is known to be quicker than shapely. 
        Default is `shapely`.
    simplify_algorithm : str
        Choose between 'dp' and 'vw', for Douglas-Peucker or Visvalingam-Whyatt 
        respectively. 'vw' will only be selected if `simplify_with` is set to 
        `simplification`. Default is `dp`, since it still "produces the most accurate 
        generalization" (Chi & Cheung, 2006).
    winding_order : str
        Determines the winding order of the features in the output geometry. Choose 
        between `CW_CCW` for clockwise orientation for outer rings and counter-clockwise
        for interior rings. Or `CCW_CW` for counter-clockwise for outer rings and 
        clockwise for interior rings. Default is `CW_CCW`.
    """

    def __init__(
        self,
        data,
        topology=True,
        prequantize=True,
        topoquantize=False,
        presimplify=False,
        toposimplify=0.0001,
        simplify_with="shapely",
        simplify_algorithm="dp",
        winding_order="CW_CCW",
    ):

        options = TopoOptions(locals())
        # execute previous steps
        super().__init__(data, options)

        # execute main function of Topology
        self.output = self.worker(self.output)

    def __repr__(self):
        return "Topology(\n{}\n)".format(pprint.pformat(self.output))

    @property
    def __geo_interface__(self):
        topo_object = copy.deepcopy(self.output)
        return serialize_as_geojson(topo_object, validate=False, lyr_idx=0)

    def to_dict(self, options=False):
        topo_object = copy.deepcopy(self.output)
        topo_object = self.resolve_coords(topo_object)
        if options is False:
            del topo_object["options"]
        elif options is True:
            topo_object["options"] = vars(topo_object["options"])
        return topo_object

    def to_svg(self, separate=False, include_junctions=False):
        serialize_as_svg(self.output, separate, include_junctions)

    def to_json(self, fp=None, options=False, pretty=False, indent=4, maxlinelength=88):
        topo_object = copy.deepcopy(self.output)
        topo_object = self.resolve_coords(topo_object)
        if options is False:
            del topo_object["options"]
        elif options is True:
            topo_object["options"] = vars(topo_object["options"])
        return serialize_as_json(
            topo_object, fp, pretty=pretty, indent=indent, maxlinelength=maxlinelength
        )

    def to_geojson(
        self,
        fp=None,
        pretty=False,
        indent=4,
        maxlinelength=88,
        validate=False,
        lyr_idx=0,
    ):
        topo_object = copy.deepcopy(self.output)
        topo_object = self.resolve_coords(topo_object)
        fc = serialize_as_geojson(topo_object, validate=validate, lyr_idx=lyr_idx)
        return serialize_as_json(
            fc, fp, pretty=pretty, indent=indent, maxlinelength=maxlinelength
        )

    def to_gdf(self):
        from ..utils import serialize_as_geodataframe

        topo_object = copy.deepcopy(self.output)
        topo_object = self.resolve_coords(topo_object)
        del topo_object["options"]
        return serialize_as_geodataframe(topo_object)

    def to_alt(
        self,
        mesh=True,
        color=None,
        tooltip=True,
        projection="identity",
        objectname="data",
    ):
        from ..utils import serialize_as_altair

        topo_object = copy.deepcopy(self.output)
        topo_object = self.resolve_coords(topo_object)
        del topo_object["options"]
        return serialize_as_altair(
            topo_object, mesh, color, tooltip, projection, objectname
        )

    def to_widget(
        self,
        slider_toposimplify={"min": 0, "max": 10, "step": 0.01, "value": 0.01},
        slider_topoquantize={"min": 1, "max": 6, "step": 1, "value": 1e5, "base": 10},
    ):

        from ..utils import serialize_as_ipywidgets

        return serialize_as_ipywidgets(
            topo_object=self,
            toposimplify=slider_toposimplify,
            topoquantize=slider_topoquantize,
        )

    def flatten_properties(self):
        objects = self.output["objects"]["data"]["geometries"]
        if objects:
            objects = properties_foreign(objects)
            self.output["objects"]["data"]["geometries"] = objects

    def topoquantize(self, quant_factor, inplace=False):
        result = copy.deepcopy(self)
        arcs = result.output["arcs"]

        # dequantize if quantization is applied
        if "transform" in result.output.keys():
            np_arcs = np_array_from_arcs(arcs)

            transform = result.output["transform"]
            scale = transform["scale"]
            translate = transform["translate"]

            np_arcs = dequantize(np_arcs, scale, translate)
            l_arcs = []
            for ls in np_arcs:
                l_arcs.append(ls[~np.isnan(ls)[:, 0]].tolist())
            arcs = l_arcs

        arcs_qnt, transform = quantize(arcs, result.output["bbox"], quant_factor)

        result.output["arcs"] = delta_encoding(arcs_qnt)
        result.output["transform"] = transform
        result.options.topoquantize = quant_factor

        if inplace:
            # update into self
            self = result
        else:
            return result

    def toposimplify(self, epsilon, _input_as="array", inplace=False):
        result = copy.deepcopy(self)

        arcs = result.output["arcs"]
        if arcs:
            np_arcs = np_array_from_arcs(arcs)

            # dequantize if quantization is applied
            if "transform" in result.output.keys():

                transform = result.output["transform"]
                scale = transform["scale"]
                translate = transform["translate"]

                np_arcs = dequantize(np_arcs, scale, translate)

            result.output["arcs"] = simplify(
                np_arcs,
                epsilon,
                algorithm=result.options.simplify_algorithm,
                package=result.options.simplify_with,
                input_as=_input_as,
            )

        # quantize aqain if quantization was applied
        if "transform" in result.output.keys():
            if result.options.topoquantize > 0:
                # set default if not specifically given in the options
                if type(result.options.topoquantize) == bool:
                    quant_factor = 1e6
                else:
                    quant_factor = result.options.topoquantize
            elif result.options.prequantize > 0:
                # set default if not specifically given in the options
                if type(result.options.prequantize) == bool:
                    quant_factor = 1e6
                else:
                    quant_factor = result.options.prequantize

            result.output["arcs"], transform = quantize(
                result.output["arcs"], result.output["bbox"], quant_factor
            )

            result.output["coordinates"], transform = quantize(
                result.output["coordinates"], result.output["bbox"], quant_factor
            )

            result.output["arcs"] = delta_encoding(result.output["arcs"])
            result.output["transform"] = transform
        if inplace:
            # update into self
            self.output["arcs"] = result.output["arcs"]
            if "transform" in result.output.keys():
                self.output["transform"] = result.output["transform"]
            # self.output["arcs"] = result.output["arcs"]
            # self.output["transform"] = result.output["transform"]
        else:
            return result

    def resolve_coords(self, data):
        geoms = data["objects"]["data"]["geometries"]
        for idx, feat in enumerate(geoms):
            if feat["type"] in ["Point", "MultiPoint"]:

                lofl = feat["coordinates"]
                repeat = 1 if feat["type"] == "Point" else 2

                for _ in range(repeat):
                    lofl = list(itertools.chain(*lofl))

                for idx, val in enumerate(lofl):
                    coord = data["coordinates"][val]
                    lofl[idx] = [int(coord.xy[0][0]), int(coord.xy[1][0])]

                feat["coordinates"] = lofl[0] if len(lofl) == 1 else lofl
                feat.pop("reset_coords", None)
        data.pop("coordinates", None)
        return data

    def worker(self, data):
        self.output["arcs"] = data["linestrings"]
        del data["linestrings"]

        # apply delta-encoding if prequantization is applied
        if self.options.prequantize > 0:
            self.output["arcs"] = delta_encoding(self.output["arcs"])
        else:
            for idx, ls in enumerate(self.output["arcs"]):
                self.output["arcs"][idx] = np.array(ls).tolist()

        # toposimplify linestrings if required
        if self.options.toposimplify > 0:
            # set default if not specifically given in the options
            if type(self.options.toposimplify) == bool:
                simplify_factor = 0.0001
            else:
                simplify_factor = self.options.toposimplify

            self.toposimplify(epsilon=simplify_factor, _input_as="array", inplace=True)

        return self.output



import topojson
import geopandas
from shapely import geometry
from shapely import wkt
from topojson.core.dedup import Dedup


# duplicate rotated geometry bar with hole interior in geometry foo
def test_dedup_duplicate_rotated_hole_interior():
    data = {
        "foo": {
            "type": "MultiPolygon",
            "coordinates": [
                [
                    [[0, 0], [20, 0], [10, 20], [0, 0]],  # CCW
                    [[3, 2], [10, 16], [17, 2], [3, 2]],  # CW
                ],
                [[[6, 4], [14, 4], [10, 12], [6, 4]]],  # CCW
            ],
        },
        "bar": {
            "type": "Polygon",
            "coordinates": [[[17, 2], [3, 2], [10, 16], [17, 2]]],
        },
    }
    topo = Dedup(data).to_dict()

    assert len(topo["bookkeeping_duplicates"]) == 0
    assert topo["bookkeeping_geoms"] == [[0, 1], [2], [3]]


def test_dedup_two_polygon_reversed_shared_arc():
    data = {
        "abcda": {
            "type": "Polygon",
            "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]],
        },  # rotated to BCDAB, cut BC-CDAB
        "befcb": {
            "type": "Polygon",
            "coordinates": [[[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]],
        },
    }
    topo = Dedup(data).to_dict()

    assert len(topo["bookkeeping_duplicates"]) == 0
    assert topo["bookkeeping_shared_arcs"] == [2]
    assert topo["bookkeeping_arcs"] == [[2, 0], [1, 2]]


def test_dedup_duplicate_polygon_no_junctions():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]]},
        "acba": {"type": "Polygon", "coordinates": [[[0, 0], [0, 1], [1, 0], [0, 0]]]},
    }
    topo = Dedup(data).to_dict()

    assert len(topo["bookkeeping_duplicates"]) == 0
    assert topo["bookkeeping_shared_arcs"] == [0]
    assert topo["bookkeeping_arcs"] == [[0], [0]]
    assert topo["bookkeeping_geoms"] == [[0], [1]]


def test_dedup_shared_line_ABCDBE_and_FBCG():
    data = {
        "abcdbe": {
            "type": "LineString",
            "coordinates": [[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]],
        },
        "fbcg": {"type": "LineString", "coordinates": [[0, 1], [1, 0], [2, 0], [3, 1]]},
    }
    topo = Dedup(data).to_dict()

    assert len(topo["bookkeeping_duplicates"]) == 0
    assert topo["bookkeeping_shared_arcs"] == [4]
    assert topo["bookkeeping_arcs"] == [[0, 4, 1, 2], [3, 4, 5]]


# this test was added since the shared_arcs bookkeeping is not doing well. Next runs
# can affect previous runs where dup_pair_list should update properly.
def test_dedup_shared_junctions_in_shared_paths():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[
        (data.name == "Togo") | (data.name == "Benin") | (data.name == "Burkina Faso")
    ]
    topo = Dedup(data).to_dict()

    assert len(topo["bookkeeping_shared_arcs"]) == 3
    assert len(topo["bookkeeping_duplicates"]) == 0


# this test was added since the shared_arcs bookkeeping is doing well, but the
# wrong arc gots deleted. How come?
# the problem arose in wrongly linemerging of contigiuous line-elements
# merged linestring is not always placed upfront.
def test_dedup_arc_not_shared_arcs_got_deleted():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[
        (data.name == "Botswana")
        | (data.name == "South Africa")
        | (data.name == "Zimbabwe")
        | (data.name == "Mozambique")
        | (data.name == "Zambia")
    ]
    topo = Dedup(data).to_dict()

    assert len(topo["bookkeeping_shared_arcs"]) == 9
    assert len(topo["bookkeeping_duplicates"]) == 0


# this test was added since there is no test for non-intersecting geometries.
# as was raised in https://github.com/mattijn/topojson/issues/1
def test_dedup_no_shared_paths_in_geoms():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.name == "Togo") | (data.name == "Liberia")]
    topo = Dedup(data).to_dict()

    assert len(topo["bookkeeping_shared_arcs"]) == 0
    assert len(topo["bookkeeping_duplicates"]) == 0


# this test was added since the local variable "array_bk_sarcs" was
# referenced before assignment. As was raised in:
# https://github.com/mattijn/topojson/issues/3
def test_dedup_array_bk_sarcs_reference():
    data = {
        "foo": {"type": "LineString", "coordinates": [[4, 0], [2, 2], [0, 0]]},
        "bar": {
            "type": "LineString",
            "coordinates": [[0, 2], [1, 1], [2, 2], [3, 1], [4, 2]],
        },
    }
    topo = Dedup(data).to_dict()

    assert len(topo["bookkeeping_shared_arcs"]) == 1
    assert len(topo["junctions"]) == 4


def test_dedup_super_function():
    data = geometry.GeometryCollection(
        [
            geometry.Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]),
            geometry.Polygon([[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]),
        ]
    )
    topo = Dedup(data).to_dict()

    assert list(topo.keys()) == [
        "type",
        "linestrings",
        "coordinates",
        "bookkeeping_geoms",
        "bookkeeping_coords",
        "objects",
        "options",
        "bbox",
        "junctions",
        "bookkeeping_duplicates",
        "bookkeeping_arcs",
        "bookkeeping_shared_arcs",
    ]


# this test was added since there is an error stating the following during Dedup:
# TypeError: list indices must be integers or slices, not NoneType, see #50
def test_dedup_s2_geometries():
    data = [
        wkt.loads(
            "MULTILINESTRING ((-51.17176115208171 -30.05269620283153, -51.18859500873385 -29.99305326146263, -51.1541142383379 -29.95234110496228, -51.13731737261026 -30.01193511071039, -51.17176115208171 -30.05269620283153), (-51.13731737261026 -30.01193511071039, -51.1541142383379 -29.95234110496228, -51.11963364027719 -29.91170657721793, -51.10287369862932 -29.97125162042611, -51.13731737261026 -30.01193511071039), (-51.13799328025614 -30.17188406207867, -51.17176115208171 -30.05269620283153, -51.10287369862932 -29.97125162042611, -51.06925390117097 -30.09024489967364, -51.13799328025614 -30.17188406207867), (-51.06925390117097 -30.09024489967364, -51.0860804353923 -30.03076444145886, -51.05167386668366 -29.99010960397871, -51.03488427131447 -30.04954147652281, -51.06925390117097 -30.09024489967364), (-51.0860804353923 -30.03076444145886, -51.10287369862932 -29.97125162042611, -51.0684302317277 -29.9306455702365, -51.05167386668366 -29.99010960397871, -51.0860804353923 -30.03076444145886))"
        )
    ]
    topo = Dedup(data).to_dict()

    assert len(topo["junctions"]) == 6
    assert len(topo["bookkeeping_duplicates"]) == 0


def test_dedup_linemerge_multilinestring():
    data = [
        {"type": "LineString", "coordinates": [(0, 0), (10, 0), (10, 5), (20, 5)]},
        {
            "type": "LineString",
            "coordinates": [
                (5, 0),
                (25, 0),
                (25, 5),
                (16, 5),
                (16, 10),
                (14, 10),
                (14, 5),
                (0, 5),
            ],
        },
    ]
    topo = Dedup(data).to_dict()

    assert len(topo["linestrings"]) == 9
    assert len(topo["junctions"]) == 7



import copy
import pprint
import numpy as np
from shapely import geometry
from shapely.ops import linemerge
from .dedup import Dedup
from ..ops import is_ccw
from ..utils import serialize_as_svg
from ..utils import serialize_as_json
from ..utils import serialize_as_altair


class Hashmap(Dedup):
    """
    hash arcs based on their type
    """

    def __init__(self, data, options={}):
        # execute previous step
        super().__init__(data, options)

        # initation topology items
        self.simplified = False

        # execute main function of Hashmap
        self.output = self.hashmapper(self.output)

    def __repr__(self):
        return "Hashmap(\n{}\n)".format(pprint.pformat(self.output))

    def to_dict(self):
        topo_object = copy.copy(self.output)
        topo_object["options"] = vars(topo_object["options"])
        return topo_object

    def to_svg(self, separate=False, include_junctions=False):
        serialize_as_svg(self.output, separate, include_junctions)

    def to_json(self, fp=None):
        topo_object = copy.copy(self.output)
        topo_object["options"] = vars(topo_object["options"])
        return serialize_as_json(topo_object, fp)

    def to_alt(
        self,
        mesh=True,
        color=None,
        tooltip=True,
        projection="identity",
        objectname="data",
    ):
        topo_object = copy.copy(self.output)
        del topo_object["options"]
        return serialize_as_altair(
            topo_object, mesh, color, tooltip, projection, objectname
        )

    def hashmapper(self, data):
        """
        Hashmap function resolves bookkeeping results to object arcs.

        The hashmap function is the fifth step in the topology computation.
        The following sequence is adopted:
        1. extract
        2. join
        3. cut
        4. dedup
        5. hashmap

        Developping Notes:
        * PostGIS Tips for Power Users: http://2010.foss4g.org/presentations/3369.pdf
        """

        # make data available within class
        self.data = data

        # resolve bookkeeping to arcs in objects, including backward check of arcs
        # resolve bookkeeping of coordinates in objects, including delta-encodig
        list(self.resolve_objects(["arcs", "coordinates"], self.data["objects"]))

        objects = {}
        objects["geometries"] = []
        objects["type"] = "GeometryCollection"
        for feature in data["objects"]:
            feat = data["objects"][feature]

            if "geometries" in feat and len(feat["geometries"]) == 1:
                feat["type"] = feat["geometries"][0]["type"]

            self.resolve_arcs(feat)

            objects["geometries"].append(feat)

        data["objects"] = {}
        data["objects"]["data"] = objects

        # prepare to return object
        data = self.data
        del data["junctions"]
        del data["bookkeeping_geoms"]
        del data["bookkeeping_coords"]
        del data["bookkeeping_duplicates"]
        del data["bookkeeping_arcs"]
        del data["bookkeeping_shared_arcs"]

        return data

    def hash_order(self, arc_ids, shared_bool):
        """
        create a decision list with the following options:
        0 - skip the array
        1 - follow the order of the first arc
        2 - follow the order of the last arc
        3 - align first two arcs and continue

        Parameters
        ----------
        arc_ids : list
            list containing the index values of the arcs
        shared_bool : list
            boolean list with same length as arc_ids,
            True means the arc is shared, False means it is a non-shared arc

        Returns
        -------
        order_of_arc : numpy array
            array containg values if first or last arc should be used to order
        split_arc_ids : list of numpy arrays
            array containing splitted arc ids
        """

        # first split it by the boolean shared arcs
        split_arc_ids = np.split(arc_ids, np.nonzero(~shared_bool)[0])
        split_boolean = np.split(shared_bool, np.nonzero(~shared_bool)[0])

        order_of_arc = [None] * len(split_boolean)
        # no need for iterations if split_boolean contains only 1 array
        if len(split_boolean) == 1:
            # happens when the geom only contains shared arcs,  its impossible to guess
            # the direction of the geometry, so align the first two arcs and continue
            # from there
            order_of_arc[0] = 3

        else:
            # check each splitted geom to decide how to treat
            for idx, split_geom in enumerate(split_boolean):
                if len(split_geom) == 0:
                    # skip when the array is empty
                    order_of_arc[idx] = 0

                elif len(split_geom) != 0 and split_geom.sum() == 0:
                    # skip when the splitted geom contains only non-shared arcs
                    order_of_arc[idx] = 0

                elif len(split_geom) != 0 and split_geom.sum() == len(split_geom):
                    # happens when the first splitted geom contains only shared arcs
                    # but the first arc of next splitted geom contains a non-shared arc
                    # lets take that one, and follow the order of the last arc
                    next_arc = split_arc_ids[idx + 1][0]
                    split_arc_ids[idx] = np.append(split_arc_ids[idx], next_arc)
                    order_of_arc[idx] = 2

                else:
                    # len(split_geom) > 1 and split_geom.sum() != len(split_geom):
                    order_of_arc[idx] = 1

        return order_of_arc, split_arc_ids

    def backward_arcs(self, arc_ids):
        """
        Function to check if the shared arcs in geom should be backward.
        If so, are written as -(index+1)

        Parameters
        ----------
        arc_ids : list
            description of input

        Returns
        -------
        arc_ids : list
            description of output
        """

        shared_bool = np.isin(arc_ids, self.data["bookkeeping_shared_arcs"])
        order_of_arc, split_arc_ids = self.hash_order(arc_ids, shared_bool)

        for idx_outer, split_arc in enumerate(split_arc_ids):
            order = order_of_arc[idx_outer]

            # if order is 0 can skip the split_arc
            # if order is 1 should follow order of the first arc
            # if order is 2 should follow order of the last arc
            # if order is 3 align first two arcs and continue
            if order == 0:
                continue

            if order == 2:
                split_arc = split_arc[::-1]

            previous_arc_backwards = False
            for idx, arc_idx in enumerate(split_arc):
                if idx == 0:
                    continue

                # seems previous run can influence next run
                arc_idx_prev = split_arc[idx - 1]
                if arc_idx_prev < 0:
                    arc_idx_prev = abs(arc_idx_prev) - 1

                current_arc = self.data["linestrings"][arc_idx]
                previous_arc = self.data["linestrings"][arc_idx_prev]

                # get first and last coordinate of current and previous arc
                coord_f = [current_arc.xy[0][0], current_arc.xy[1][0]]
                coord_l = [current_arc.xy[0][-1], current_arc.xy[1][-1]]

                if not previous_arc_backwards:
                    coords_prev = [
                        [previous_arc.xy[0][0], previous_arc.xy[1][0]],
                        [previous_arc.xy[0][-1], previous_arc.xy[1][-1]],
                    ]
                else:
                    coords_prev = [
                        [previous_arc.xy[0][-1], previous_arc.xy[1][-1]],
                        [previous_arc.xy[0][0], previous_arc.xy[1][0]],
                    ]
                coord_f_prev = coords_prev[0]
                coord_l_prev = coords_prev[1]

                # order 1, compare last coordinate of previous arc with first coordinate
                # of current arc. If not equal, rotate current arc
                if order == 1:
                    if not np.array_equiv(coord_l_prev, coord_f):
                        split_arc[idx] = -(arc_idx + 1)
                        previous_arc_backwards = True
                    else:
                        previous_arc_backwards = False

                # order 2, since the list is reversed, have to check first coordinate
                # of previous arc with the last coordinate of current arc. If not equal
                # rotate current arc.
                elif order == 2:
                    if not np.array_equiv(coord_f_prev, coord_l):
                        split_arc[idx] = -(arc_idx + 1)
                        previous_arc_backwards = True
                    else:
                        previous_arc_backwards = False

                elif order == 3:
                    if np.array_equiv(coord_f_prev, coord_l) and not np.array_equiv(
                        coord_l_prev, coord_f
                    ):
                        split_arc[idx - 1] = -(arc_idx_prev + 1)
                        split_arc[idx] = -(arc_idx + 1)
                        previous_arc_backwards = True
                    elif np.array_equiv(coord_f, coord_f_prev):
                        split_arc[idx - 1] = -(arc_idx_prev + 1)
                        previous_arc_backwards = False
                    elif not np.array_equiv(coord_l_prev, coord_f):
                        split_arc[idx] = -(arc_idx + 1)
                        previous_arc_backwards = True
                    else:
                        previous_arc_backwards = False

            if order == 2:
                split_arc_ids[idx_outer] = split_arc[::-1]

        comb_arc_ids = np.concatenate(split_arc_ids).flatten()
        _, idx_arcs = np.unique(comb_arc_ids, return_index=True)
        arc_ids = comb_arc_ids[np.sort(idx_arcs)].tolist()

        if order == 3:
            # since alignment is done based on the first two arcs, need a double-check
            # if it follows the required order of the ring
            if self.inner and self.options.winding_order == "CCW_CW":
                need_ccw = False
            elif not self.inner and (
                self.options.winding_order == "CW_CCW"
                or self.options.winding_order is None
            ):
                need_ccw = False
            else:
                need_ccw = True

            arc_ids = self.resolve_orient(arc_ids, need_ccw)

        return arc_ids

    def resolve_orient(self, arcs_idx_geom, need_ccw):
        arcs_geom = []
        for arc_idx in arcs_idx_geom:
            if arc_idx < 0:
                arc = copy.copy(self.data["linestrings"][~arc_idx])
                arc.coords = list(arc.coords)[::-1]
                arcs_geom.append(arc)
            else:
                arc = self.data["linestrings"][arc_idx]
                arcs_geom.append(arc)
        lring = geometry.LinearRing(linemerge(arcs_geom))

        if is_ccw(lring) != need_ccw:
            arcs_idx_geom = (np.array(arcs_idx_geom) * -1 - 1).tolist()

        return arcs_idx_geom

    def resolve_bookkeeping(self, geoms, key):
        """
        Function that is activated once the key of interest in the find_arcs function
        is detected. It replaces the geom ids with the corresponding arc ids.
        """

        if key == "arcs":
            bk_objects = "bookkeeping_geoms"
            bk_element = "bookkeeping_arcs"
        elif key == "coordinates":
            bk_objects = bk_element = "bookkeeping_coords"

        arcs = []
        for geom in geoms:
            arcs_in_geom = self.data[bk_objects][geom]
            for idx_arc, arc_ref in enumerate(arcs_in_geom):
                arc_ids = self.data[bk_element][arc_ref]
                if len(arc_ids) > 1 and key is not "coordinates":
                    self.inner = True if idx_arc > 0 else False
                    arc_ids = self.backward_arcs(arc_ids)

                arcs.append(arc_ids)
        return arcs

    def resolve_objects(self, keys, dictionary):
        """
        Function that resolves the bookkeeping back to the arcs in the objects.
        Support also nested dicts such as GeometryCollections
        """

        for k, v in dictionary.items():
            # resolve when key equals 'arcs' and v contains arc indici
            if str(k) in keys and v is not None:
                dictionary[k] = self.resolve_bookkeeping(v, k)
                yield v
            elif isinstance(v, dict):
                for result in self.resolve_objects(keys, v):
                    yield result
            elif isinstance(v, list):
                for d in v:
                    for result in self.resolve_objects(keys, d):
                        yield result

    def resolve_arcs(self, feat):
        """
        Function that resolves the arcs based on the type of the feature
        """
        if feat["type"] == "LineString":
            if "geometries" in feat:
                f_arc = feat["geometries"][0]["arcs"][0]
            else:
                f_arc = feat["arcs"][0]
            feat["arcs"] = f_arc
            feat.pop("geometries", None)

        elif feat["type"] == "Polygon":
            if "geometries" in feat:
                f_arc = feat["geometries"][0]["arcs"]
            else:
                f_arc = feat["arcs"]

            feat["arcs"] = f_arc
            feat.pop("geometries", None)

        elif feat["type"] == "MultiPolygon":
            if "geometries" in feat:
                f_arcs = feat["geometries"][0]["arcs"]
            else:
                f_arcs = feat["arcs"]
            feat["arcs"] = [[arc] for arc in f_arcs]
            feat.pop("geometries", None)

        elif feat["type"] == "GeometryCollection":
            feat["geometries"] = [
                self.resolve_arcs(feat) for feat in feat["geometries"]
            ]

        elif feat["type"] == "Point":
            if "geometries" in feat:
                f_arc = feat["geometries"][0]["coordinates"]
            else:
                f_arc = feat["coordinates"]

            feat["coordinates"] = f_arc
            feat.pop("geometries", None)

        elif feat["type"] == "MultiPoint":
            if "geometries" in feat:
                f_arcs = feat["geometries"][0]["coordinates"]
            else:
                f_arcs = feat["coordinates"]
            feat["coordinates"] = [[arc] for arc in f_arcs]
            feat.pop("geometries", None)

        return feat



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['topojson', 'topojson.core']

package_data = \
{'': ['*'],
 'topojson': ['.ipynb_checkpoints/*'],
 'topojson.core': ['.ipynb_checkpoints/*']}

install_requires = \
['numpy', 'shapely']

extras_require = \
{'dev': ['shapely >=1.7a2',
         'geojson',
         'simplification',
         'pyshp',
         'fiona >=1.8.6',
         'geopandas',
         'altair',
         'ipywidgets']}

setup(name='topojson',
      version='1.0rc9',
      description='topojson - a powerful library to encode geographic data as topology in Python!🌍',
      author='Mattijn van Hoek',
      author_email='mattijn@gmail.com',
      url='https://github.com/mattijn/topojson',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )



import topojson
import geopandas
from shapely import geometry
from topojson.core.cut import Cut


# cut exact duplicate lines ABC & ABC have no cuts
def test_cut_exact_duplicate_lines_ABC_ABC_no_cuts():
    data = {
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "abc2": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
    }
    topo = Cut(data).to_dict()

    assert len(topo["junctions"]) == 0
    assert topo["bookkeeping_duplicates"].tolist() == [[1, 0]]


# cut reversed duplicate lines ABC & CBA have no cuts
def test_cut_reversed_duplicate_lines_ABC_CBA_no_cuts():
    data = {
        "abc": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
        "cba": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
    }
    topo = Cut(data).to_dict()

    assert len(topo["junctions"]) == 0
    assert topo["bookkeeping_duplicates"].tolist() == [[1, 0]]


# cut exact duplicate rings ABCA & ABCA have no cuts
def test_cut_exact_duplicate_rings_ABCA_ABCA_no_cuts():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [2, 1], [0, 0]]]},
        "abca2": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [2, 1], [0, 0]]]},
    }
    topo = Cut(data).to_dict()

    assert len(topo["junctions"]) == 0
    assert topo["bookkeeping_duplicates"].tolist() == [[1, 0]]


# cut reversed rings ABCA & ACBA have no cuts
def test_cut_reversed_rings_ABCA_ACBA_no_cuts():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [2, 1], [0, 0]]]},
        "acba": {"type": "Polygon", "coordinates": [[[0, 0], [2, 1], [1, 0], [0, 0]]]},
    }
    topo = Cut(data).to_dict()

    assert len(topo["junctions"]) == 0
    assert topo["bookkeeping_duplicates"].tolist() == [[1, 0]]


# cut rotated duplicate rings BCAB & ABCA have no cuts
def test_cut_rotated_duplicates_rings_BCAB_ABCA_no_cuts():
    data = {
        "abca": {"type": "Polygon", "coordinates": [[[0, 0], [1, 0], [2, 1], [0, 0]]]},
        "bcab": {"type": "Polygon", "coordinates": [[[1, 0], [2, 1], [0, 0], [1, 0]]]},
    }
    topo = Cut(data).to_dict()

    assert len(topo["junctions"]) == 0
    assert topo["bookkeeping_duplicates"].tolist() == [[1, 0]]


# cut ring ABCA & line ABCA have no cuts
def test_cut_ring_ABCA_line_ABCA_no_cuts():
    data = {
        "abcaLine": {
            "type": "Linestring",
            "coordinates": [[0, 0], [1, 0], [2, 1], [0, 0]],
        },
        "abcaPolygon": {
            "type": "Polygon",
            "coordinates": [[[0, 0], [1, 0], [2, 1], [0, 0]]],
        },
    }
    topo = Cut(data).to_dict()

    assert len(topo["junctions"]) == 0
    assert topo["bookkeeping_duplicates"].tolist() == [[1, 0]]


# cut ring BCAB & line ABCA have no cuts
def test_cut_ring_BCAB_line_ABCA_no_cuts():
    data = {
        "abcaLine": {
            "type": "Linestring",
            "coordinates": [[0, 0], [1, 0], [2, 1], [0, 0]],
        },
        "bcabPolygon": {
            "type": "Polygon",
            "coordinates": [[[1, 0], [2, 1], [0, 0], [1, 0]]],
        },
    }
    topo = Cut(data).to_dict()

    assert len(topo["junctions"]) == 0
    assert topo["bookkeeping_duplicates"].tolist() == [[1, 0]]


# cut ring ABCA & line BCAB have no cuts
def test_cut_ring_ABCA_line_BCAB_no_cuts():
    data = {
        "bcabLine": {
            "type": "Linestring",
            "coordinates": [[1, 0], [2, 1], [0, 0], [1, 0]],
        },
        "abcaPolygon": {
            "type": "Polygon",
            "coordinates": [[[0, 0], [1, 0], [2, 1], [0, 0]]],
        },
    }
    topo = Cut(data).to_dict()

    assert len(topo["junctions"]) == 0
    assert topo["bookkeeping_duplicates"].tolist() == [[1, 0]]


# overlapping rings ABCDA and BEFCB are cut into BC-CDAB and BEFC-CB
def test_cut_overlapping_rings_are_cut():
    data = {
        "abcda": {
            "type": "Polygon",
            "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]],
        },  # rotated to BCDAB, cut BC-CDAB
        "befcb": {
            "type": "Polygon",
            "coordinates": [[[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]],
        },
    }
    topo = Cut(data).to_dict()

    assert topo["bookkeeping_linestrings"].size == 6
    assert topo["bookkeeping_duplicates"].tolist() == [[4, 1]]


# currently the border between Sudan and Eqypt is not recognized as duplicate
# because of floating-precision. Should be solved by a "snap_to_grid" option.
def test_cut_border_egypt_sudan():
    data = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    data = data[(data.name == "Egypt") | (data.name == "Sudan")]
    topo = Cut(data).to_dict()

    assert len(topo["bookkeeping_duplicates"].tolist()) == 1


def test_cut_nybb_fast_split():
    nybb_path = geopandas.datasets.get_path("nybb")
    data = geopandas.read_file(nybb_path)
    data.set_index("BoroCode", inplace=True)
    topo = Cut(data).to_dict()

    assert topo["bookkeeping_linestrings"].size == 5618


# this test was added since the fast_split was really slow on geometries
# when there are many junctions in the geometry. During debugging this test ran
# eventually 8x more quick.
def test_cut_many_junctions():
    data = geopandas.read_file("tests/files_geojson/mesh2d.geojson", driver="GeoJSON")
    # previous test ran in 8.798s (best of 3)
    # current test ran in 8.182s (best of 3)
    topo = Cut(data).to_dict()

    assert topo["bookkeeping_linestrings"].size == 11010


def test_cut_super_function_cut():
    data = geometry.GeometryCollection(
        [
            geometry.Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]),
            geometry.Polygon([[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]),
        ]
    )
    topo = Cut(data).to_dict()

    assert list(topo.keys()) == [
        "type",
        "linestrings",
        "coordinates",
        "bookkeeping_geoms",
        "bookkeeping_coords",
        "objects",
        "options",
        "bbox",
        "junctions",
        "bookkeeping_duplicates",
        "bookkeeping_linestrings",
    ]


# this geometry is added on several classes (extract and hashmap). Not yet clear in
# which phase it is going wrong
def test_cut_geomcol_multipolygon_polygon():
    data = {
        "foo": {
            "type": "GeometryCollection",
            "geometries": [
                {
                    "type": "MultiPolygon",
                    "coordinates": [
                        [
                            [[10, 20], [20, 0], [0, 0], [10, 20]],
                            [[3, 2], [10, 16], [17, 2], [3, 2]],
                        ],
                        [[[6, 4], [14, 4], [10, 12], [6, 4]]],
                    ],
                },
                {
                    "type": "Polygon",
                    "coordinates": [[[20, 0], [35, 5], [10, 20], [20, 0]]],
                },
            ],
        }
    }
    topo = Cut(data).to_dict()

    assert topo["bookkeeping_linestrings"].size == 8


# this test is added since its seems no extra junctions are placed at lines where other
# linestrings have shared paths but no shared junctions.
def test_cut_linemerge_multilinestring():
    data = [
        {"type": "LineString", "coordinates": [(0, 0), (10, 0), (10, 5), (20, 5)]},
        {
            "type": "LineString",
            "coordinates": [
                (5, 0),
                (25, 0),
                (25, 5),
                (16, 5),
                (16, 10),
                (14, 10),
                (14, 5),
                (0, 5),
            ],
        },
    ]
    topo = Cut(data).to_dict()

    assert len(topo["linestrings"]) == 12
    assert len(topo["junctions"]) == 7
